#!/usr/bin/env python
import os
import datetime
import yaml
from napa.utils.serials import *
from napa.utils.io import *

class Config():

    def __init__(self, path, command_type):
        '''
        Generate run configuration object from 
        user-provided config file in yaml format
        '''
        with open(path, 'r') as f:
            config = yaml.load(f)
        
        self.__dict__.update(config)
        self.set_run_type(command_type)
        self.check_input()

    #---------------------------------------------------------#
    def set_run_type(self, command_type):
        self.build_network = False
        self.analyze_network = False
        
        if 'build' in command_type:
            self.build_network = True
        
        if 'analy' in command_type:
            self.analyze_network = True

    #---------------------------------------------------------#
    def check_input(self):
        '''
        Checks input for consistently defined parameters.
        '''

        # Check main command and parameter options
        self.check_param_list(['net_type', 'edge_type'],
                              'Required command parameters')

        # Missing directories
        self.check_param_list(['working_dir', 'data_dir', 
                               'results_dir'], 
                              'Required Directories')
        self.set_net_file_base()
        self.set_net_file_name()

        # Check network reconstruction required parameters
        if self.build_network:
            self.build_net_input()

        # Check network analysis required parameters
        if self.analyze_network:
            self.analyze_net_input()

    #---------------------------------------------------------#
    def set_net_file_base(self):
        if not hasattr(self, 'net_file_base') and \
           not hasattr(self, 'net_file'):

            stderr_write(['No network name/base given.'])
            self.net_file_base = \
                str(datetime.datetime.now()).replace(' ',
                                        '_').replace(':','-')

            if hasattr(self, 'aln_fasta_file'):
                netname = os.path.basename(self.aln_fasta_file)
                netname = os.path.splitext(netname)[0]

            self.net_file_base += '_' + netname

            stderr_write(['Set network base name to:',
                          self.net_file_base])

    #---------------------------------------------------------#
    def set_net_file_name(self):
        if self.net_file == None or not \
           to_bool(self.net_file):

            suffix = '-' + self.net_type
            method_str = '-m_'
            if hasattr(self, 'method'):
                method_str += self.method
            else:
                method_str += 'na'
            suffix += method_str 

            thresh_str = '-t_'
            if not hasattr(self, 'thresh'):
                stderr_write(['WARN: No edge weight',
                    'threshold specified.',
                    'Setting threshold to default for',
                    'mutation coocurrence counting',
                    'method (method argument).'])
                if 'jaccard' in method_str:
                    self.thresh = 0.01
                    thresh_str += str(self.thresh)                                  
                elif 'raw' in method_str:
                    self.thresh = 1
                    thresh_str += str(self.thresh)
                else:
                    thresh_str += 'na'
                
            else:
                thresh_str += str(self.thresh)
            suffix += thresh_str 
                
            if self.net_type == 'aln':
                occur_str = '-mo_'
                if hasattr(self, 'min_co_occur'):
                    occur_str += str(self.min_co_occur)
                suffix += occur_str 

            if self.net_type == 'phylo':

                edge_type_str = '-et_'
                if hasattr(self, 'edge_type'):
                    edge_type_str += self.edge_type
                suffix += edge_type_str

                dist_thresh_str = '-d_'
                if hasattr(self, 'dist_thresh'):
                    dist_thresh_str += \
                    str(self.dist_thresh)

                suffix += dist_thresh_str
                
                tree_support_str = '-mintrees_'
                if hasattr(self, 'tree_support_thresh'):
                    tree_support_str += \
                    str(self.tree_support_thresh)
                
                suffix += tree_support_str
            
            self.net_file = os.path.join(self.working_dir, 
                self.results_dir, self.net_file_base + \
                suffix + '.net.txt')

    #---------------------------------------------------------#
    def check_param_list(self, param_list, error_prefix = ''):
        '''
        Checks that all parameters in param_list are present
        and non-empty and first prints all missing parameters
        then throws a ValueError
        '''
        required_missing = False

        for param in param_list:
            if not hasattr(self, param):
                stderr_write(['Required parameter missing',
                              'from config:', param])
                required_missing = True
            elif self.__dict__[param] == None:
                stderr_write(['Empty value for required',
                              'parameter:', param])
                required_missing = True

        if required_missing:
            error_str = error_prefix 
            error_str += " ERROR: Some of these required "
            error_str += "parameters are missing or empty: "
            error_str += ' '.join(param_list)
            raise ValueError(error_str)
    #---------------------------------------------------------#
    def build_net_input(self):
        '''
        Checks and prepares input for building network 
        from alignment or phylogeny
        '''

        # REQUIRED for network construction 
        self.check_param_list(['method', 'output_edge_weight_table', 
                               'wild_type', 'aln_fasta_file'], 
                              'Net Construct Input')
        
        if to_bool(self.output_edge_weight_table):
            self.net_table_file = \
            self.net_file.replace('.net.txt', '.table.txt')
                                  
        
        # REQUIRED: FASTA format alignment for alignment 
        # and phylo networks
        self.aln_fasta_file = \
            os.path.join(self.working_dir, self.data_dir,
                         self.aln_fasta_file)
        if not os.path.isfile(self.aln_fasta_file):
            raise ValueError('Please provide a ' + \
                             'valid FASTA file (path)!\n' + \
                             self.aln_fasta_file + \
                             'does not exist!')


        # OPTIONAL: List of positions/column numbers in alignment
        # Full path to position list file if provided
        if hasattr(self, 'pos_list_file') and \
           self.pos_list_file is not None:
            self.pos_list_file = \
                os.path.join(self.working_dir, self.data_dir,
                             self.pos_list_file)
        else:
            self.pos_list_file = ''

        # OPTIONAL: File with list of protein residue 
        # positions considered for network (column format)
        # (remove residues from signaling peptide, and/or
        # other regions not studied).
        if  hasattr(self, 'pos_subset_file') \
            and self.pos_subset_file is not None:
            self.pos_subset_file = \
                os.path.join(self.working_dir, self.data_dir,
                             self.pos_subset_file)
        else:
            self.pos_subset_file = ''

        # OPTIONAL: Ranges of protein residues to include
        # Can be used instead of pos_subset_file
        if not hasattr(self, 'protein_ranges') or not \
           to_bool(self.protein_ranges):
            self.protein_ranges = []
        
        # OPTIONAL: When using protein functional selection
        # for alignment or phylo leaf sequences, we need 
        # sequence functional annotation and 
        # list of selected functions.
        if hasattr(self, 'prot_func_file') and \
           hasattr(self, 'sel_prot_func_file'):
            if self.sel_prot_func_file != None and \
               self.prot_func_file == None:

                raise ValueError(' '.join([\
                    'Protein functional assignments needed',
                    'when selecting a subset by function.']))

            if self.sel_prot_func_file != None:
                self.sel_prot_func_file = \
                    os.path.join(self.working_dir,
                                 self.data_dir,
                                 self.sel_prot_func_file)
            else:
                self.sel_prot_func_file = ''

            if len(self.sel_prot_func_file): 
                if not os.path.isfile(self.sel_prot_func_file):
                    raise ValueError(' '.join([\
                        'If you are using the selected',
                        'protein function option',
                        'please provide a valid',
                        'selected prot func. file (path).']))
                
                self.prot_func_file = \
                    os.path.join(self.working_dir, self.data_dir,
                                 self.prot_func_file)
            
                if not os.path.isfile(self.prot_func_file):
                    raise ValueError(\
                        'Please provide a valid protein ' + \
                        'functional annotation file (path).')

        # REQUIRED when constructing the phylo network
        if 'phylo' in self.net_type:
            self.check_param_list(\
                ['dist_thresh', 'tree_support_thresh',
                 'tree_file_prefix_list',
                 'tree_nwk_file_suffix',
                 'tree_internal_node_seq_suffix'], 
                 'PhyloNet Construct Input')
            self.build_phylo_net_input()

    #---------------------------------------------------------#
    def build_phylo_net_input(self):
        '''
        Check and assign input dirs/params/opts for
        reconstructing phylogeny-based network
        '''
        # Functional transitions between phylogeny sequences 
        # allowed (evolution of specific function)
        self.func_transitions = {}

        if hasattr(self, 'func_transitions_file'):
            if self.func_transitions_file != None:
                self.func_transitions_file = \
                    os.path.join(self.working_dir, 
                                 self.data_dir,
                                 self.func_transitions_file)

                if not hasattr(self,
                        'tree_internal_node_state_suffix'):
                    raise ValueError(' '.join([\
                        'Internal node functions needed when',
                        'constraining by',
                        'functional transitions.']))
        else:
            self.func_transitions_file = ''

        
        if not len(self.tree_file_prefix_list):
           raise ValueError(' '.join([\
                'PhyloNet Build Input: Invalid newick tree'
                'directory \nand/or wrong tree name prefix!']))
    
        self.dist_thresh = get_float_from_str(self.dist_thresh, 
                                              -1.)
        if self.dist_thresh < 0.:
            stderr_write(['WARNING: Distance threshold',
                          'was not given.',
                          '\nNo distance constraints for pairs',
                          'of mutations will be applied.'])
        else:
            stderr_write(['Phylogeny distance threshold set to',
                          self.dist_thresh])

        # A mutation pair has to be significantly associated 
        # in at least this many ensemble trees, default = 2
        self.tree_support_thresh = \
                get_int_from_str(self.tree_support_thresh, 2)

    #---------------------------------------------------------#
    def analyze_net_input(self):
        if not hasattr(self, 'net_file'):
             raise ValueError(\
                    'Must provide network file when ' + \
                    'analyzing network!')

        if not os.path.isfile(self.net_file):
            raise ValueError(\
                    'Network file does not exist ' + \
                    'or is not a regular file:\n' + \
                    self.net_file)

        self.check_param_list(\
            ['cluster_nodes', 'calculate_centralities'], 
            'Required for netw. analysis')
        
        if not hasattr(self, 'cent_type') or \
           self.cent_type == None:
            self.cent_type = 'loc'

        if not hasattr(self, 'partition_types') or \
           self.partition_types == None or \
           not len(self.partition_types):
            self.partition_types = ['MLV']


    #---------------------------------------------------------#
    def print_input_summary(self):
        stderr_write(['\nINITIAL INPUT PARAMETER SUMMARY:'])

        if to_bool(self.build_network):
            stderr_write(['Network will be reconstructed.'])
                
        if to_bool(self.analyze_network):
            stderr_write(['Network will be analyzed.'])

        stderr_write(['Network is', self.net_type, 'based.'])

        stderr_write(['Tab-delimited network file path:\n' + \
                      self.net_file])
        
        

        



# -*- coding: utf-8 -*-
from __future__ import print_function
'''
Copyright (c) 2009, Thomas Aynaud <thomas.aynaud@lip6.fr>
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions 
are met:

* Redistributions of source code must retain the above copyright 
notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright 
notice, this list of conditions and the following disclaimer in the 
documentation and/or other materials provided with the distribution.

* Neither the name of the NetworkX Developers nor the names of its 
contributors may be used to endorse or promote products derived from 
this software without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
'''

"""
This module implements community detection.
"""

__author__ = """Thomas Aynaud (thomas.aynaud@lip6.fr)"""
#    Copyright (C) 2009 by
#    Thomas Aynaud <thomas.aynaud@lip6.fr>
#    All rights reserved.
#    BSD license.

__PASS_MAX = -1
__MIN = 0.0000001
import array
import random

import networkx as nx

#from .community_status import Status

class Status(object):
    """
    To handle several data in one struct.
    Could be replaced by named tuple, but don't want to depend on python 2.6
    """
    node2com = {}
    total_weight = 0
    internals = {}
    degrees = {}
    gdegrees = {}

    def __init__(self):
        self.node2com = dict([])
        self.total_weight = 0
        self.degrees = dict([])
        self.gdegrees = dict([])
        self.internals = dict([])
        self.loops = dict([])

    def __str__(self):
        return ("node2com : " + str(self.node2com) + " degrees : "
                + str(self.degrees) + " internals : " + str(self.internals)
                + " total_weight : " + str(self.total_weight))

    def copy(self):
        """Perform a deep copy of status"""
        new_status = Status()
        new_status.node2com = self.node2com.copy()
        new_status.internals = self.internals.copy()
        new_status.degrees = self.degrees.copy()
        new_status.gdegrees = self.gdegrees.copy()
        new_status.total_weight = self.total_weight

    def init(self, graph, weight, part=None):
        """Initialize the status of a graph with every node in one community"""
        count = 0
        self.node2com = dict([])
        self.total_weight = 0
        self.degrees = dict([])
        self.gdegrees = dict([])
        self.internals = dict([])
        self.total_weight = graph.size(weight=weight)
        if part is None:
            for node in graph.nodes():
                self.node2com[node] = count
                deg = float(graph.degree(node, weight=weight))
                if deg < 0:
                    error = "Bad node degree ({})".format(deg)
                    raise ValueError(error)
                self.degrees[count] = deg
                self.gdegrees[node] = deg
                edge_data = graph.get_edge_data(node, node, default={weight: 0})
                self.loops[node] = float(edge_data.get(weight, 1))
                self.internals[count] = self.loops[node]
                count += 1
        else:
            for node in graph.nodes():
                com = part[node]
                self.node2com[node] = com
                deg = float(graph.degree(node, weight=weight))
                self.degrees[com] = self.degrees.get(com, 0) + deg
                self.gdegrees[node] = deg
                inc = 0.
                for neighbor, datas in graph[node].items():
                    edge_weight = datas.get(weight, 1)
                    if edge_weight <= 0:
                        error = "Bad graph type ({})".format(type(graph))
                        raise ValueError(error)
                    if part[neighbor] == com:
                        if neighbor == node:
                            inc += float(edge_weight)
                        else:
                            inc += float(edge_weight) / 2.
                self.internals[com] = self.internals.get(com, 0) + inc

def partition_at_level(dendrogram, level):
    """Return the partition of the nodes at the given level
    A dendrogram is a tree and each level is a partition of the graph nodes.
    Level 0 is the first partition, which contains the smallest communities,
    and the best is len(dendrogram) - 1.
    The higher the level is, the bigger are the communities
    Parameters
    ----------
    dendrogram : list of dict
       a list of partitions, ie dictionnaries where keys of the i+1 are the
       values of the i.
    level : int
       the level which belongs to [0..len(dendrogram)-1]
    Returns
    -------
    partition : dictionnary
       A dictionary where keys are the nodes and the values are the set it
       belongs to
    Raises
    ------
    KeyError
       If the dendrogram is not well formed or the level is too high
    See Also
    --------
    best_partition which directly combines partition_at_level and
    generate_dendrogram to obtain the partition of highest modularity
    Examples
    --------
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> dendrogram = generate_dendrogram(G)
    >>> for level in range(len(dendrogram) - 1) :
    >>>     print("partition at level", level, "is", partition_at_level(dendrogram, level))  # NOQA
    """
    partition = dendrogram[0].copy()
    for index in range(1, level + 1):
        for node, community in partition.items():
            partition[node] = dendrogram[index][community]
    return partition


def modularity(partition, graph, weight='weight'):
    """Compute the modularity of a partition of a graph
    Parameters
    ----------
    partition : dict
       the partition of the nodes, i.e a dictionary where keys are their nodes
       and values the communities
    graph : networkx.Graph
       the networkx graph which is decomposed
    weight : str, optional
        the key in graph to use as weight. Default to 'weight'
    Returns
    -------
    modularity : float
       The modularity
    Raises
    ------
    KeyError
       If the partition is not a partition of all graph nodes
    ValueError
        If the graph has no link
    TypeError
        If graph is not a networkx.Graph
    References
    ----------
    .. 1. Newman, M.E.J. & Girvan, M. Finding and evaluating community
    structure in networks. Physical Review E 69, 26113(2004).
    Examples
    --------
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> part = best_partition(G)
    >>> modularity(part, G)
    """
    if graph.is_directed():
        raise TypeError("Bad graph type, use only non directed graph")

    inc = dict([])
    deg = dict([])
    links = graph.size(weight=weight)
    if links == 0:
        raise ValueError("A graph without link has an undefined modularity")

    for node in graph:
        com = partition[node]
        deg[com] = deg.get(com, 0.) + graph.degree(node, weight=weight)
        for neighbor, datas in graph[node].items():
            edge_weight = datas.get(weight, 1)
            if partition[neighbor] == com:
                if neighbor == node:
                    inc[com] = inc.get(com, 0.) + float(edge_weight)
                else:
                    inc[com] = inc.get(com, 0.) + float(edge_weight) / 2.

    res = 0.
    for com in set(partition.values()):
        res += (inc.get(com, 0.) / links) - \
               (deg.get(com, 0.) / (2. * links)) ** 2
    return res


def best_partition(graph, partition=None,
                   weight='weight', resolution=1., randomize=False):
    """Compute the partition of the graph nodes which maximises the modularity
    (or try..) using the Louvain heuristices
    This is the partition of highest modularity, i.e. the highest partition
    of the dendrogram generated by the Louvain algorithm.
    Parameters
    ----------
    graph : networkx.Graph
       the networkx graph which is decomposed
    partition : dict, optional
       the algorithm will start using this partition of the nodes.
       It's a dictionary where keys are their nodes and values the communities
    weight : str, optional
        the key in graph to use as weight. Default to 'weight'
    resolution :  double, optional
        Will change the size of the communities, default to 1.
        represents the time described in
        "Laplacian Dynamics and Multiscale Modular Structure in Networks",
        R. Lambiotte, J.-C. Delvenne, M. Barahona
    randomize :  boolean, optional
        Will randomize the node evaluation order and the community evaluation
        order to get different partitions at each call
    Returns
    -------
    partition : dictionnary
       The partition, with communities numbered from 0 to number of communities
    Raises
    ------
    NetworkXError
       If the graph is not Eulerian.
    See Also
    --------
    generate_dendrogram to obtain all the decompositions levels
    Notes
    -----
    Uses Louvain algorithm
    References
    ----------
    .. 1. Blondel, V.D. et al. Fast unfolding of communities in
    large networks. J. Stat. Mech 10008, 1-12(2008).
    Examples
    --------
    >>>  #Basic usage
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> part = best_partition(G)
    >>> #other example to display a graph with its community :
    >>> #better with karate_graph() as defined in networkx examples
    >>> #erdos renyi don't have true community structure
    >>> G = nx.erdos_renyi_graph(30, 0.05)
    >>> #first compute the best partition
    >>> partition = community.best_partition(G)
    >>>  #drawing
    >>> size = float(len(set(partition.values())))
    >>> pos = nx.spring_layout(G)
    >>> count = 0.
    >>> for com in set(partition.values()) :
    >>>     count += 1.
    >>>     list_nodes = [nodes for nodes in partition.keys()
    >>>                                 if partition[nodes] == com]
    >>>     nx.draw_networkx_nodes(G, pos, list_nodes, node_size = 20,
                                    node_color = str(count / size))
    >>> nx.draw_networkx_edges(G, pos, alpha=0.5)
    >>> plt.show()
    """
    dendo = generate_dendrogram(graph,
                                partition,
                                weight,
                                resolution,
                                randomize)
    return partition_at_level(dendo, len(dendo) - 1)


def generate_dendrogram(graph,
                        part_init=None,
                        weight='weight',
                        resolution=1.,
                        randomize=False):
    """Find communities in the graph and return the associated dendrogram
    A dendrogram is a tree and each level is a partition of the graph nodes.
    Level 0 is the first partition, which contains the smallest communities,
    and the best is len(dendrogram) - 1. The higher the level is, the bigger
    are the communities
    Parameters
    ----------
    graph : networkx.Graph
        the networkx graph which will be decomposed
    part_init : dict, optional
        the algorithm will start using this partition of the nodes. It's a
        dictionary where keys are their nodes and values the communities
    weight : str, optional
        the key in graph to use as weight. Default to 'weight'
    resolution :  double, optional
        Will change the size of the communities, default to 1.
        represents the time described in
        "Laplacian Dynamics and Multiscale Modular Structure in Networks",
        R. Lambiotte, J.-C. Delvenne, M. Barahona
    Returns
    -------
    dendrogram : list of dictionaries
        a list of partitions, ie dictionnaries where keys of the i+1 are the
        values of the i. and where keys of the first are the nodes of graph
    Raises
    ------
    TypeError
        If the graph is not a networkx.Graph
    See Also
    --------
    best_partition
    Notes
    -----
    Uses Louvain algorithm
    References
    ----------
    .. 1. Blondel, V.D. et al. Fast unfolding of communities in large
    networks. J. Stat. Mech 10008, 1-12(2008).
    Examples
    --------
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> dendo = generate_dendrogram(G)
    >>> for level in range(len(dendo) - 1) :
    >>>     print("partition at level", level,
    >>>           "is", partition_at_level(dendo, level))
    :param weight:
    :type weight:
    """
    if graph.is_directed():
        raise TypeError("Bad graph type, use only non directed graph")

    # special case, when there is no link
    # the best partition is everyone in its community
    if graph.number_of_edges() == 0:
        part = dict([])
        for node in graph.nodes():
            part[node] = node
        return [part]

    current_graph = graph.copy()
    status = Status()
    status.init(current_graph, weight, part_init)
    status_list = list()
    __one_level(current_graph, status, weight, resolution, randomize)
    new_mod = __modularity(status)
    partition = __renumber(status.node2com)
    status_list.append(partition)
    mod = new_mod
    current_graph = induced_graph(partition, current_graph, weight)
    status.init(current_graph, weight)

    while True:
        __one_level(current_graph, status, weight, resolution, randomize)
        new_mod = __modularity(status)
        if new_mod - mod < __MIN:
            break
        partition = __renumber(status.node2com)
        status_list.append(partition)
        mod = new_mod
        current_graph = induced_graph(partition, current_graph, weight)
        status.init(current_graph, weight)
    return status_list[:]


def induced_graph(partition, graph, weight="weight"):
    """Produce the graph where nodes are the communities
    there is a link of weight w between communities if the sum of the weights
    of the links between their elements is w
    Parameters
    ----------
    partition : dict
       a dictionary where keys are graph nodes and  values the part the node
       belongs to
    graph : networkx.Graph
        the initial graph
    weight : str, optional
        the key in graph to use as weight. Default to 'weight'
    Returns
    -------
    g : networkx.Graph
       a networkx graph where nodes are the parts
    Examples
    --------
    >>> n = 5
    >>> g = nx.complete_graph(2*n)
    >>> part = dict([])
    >>> for node in g.nodes() :
    >>>     part[node] = node % 2
    >>> ind = induced_graph(part, g)
    >>> goal = nx.Graph()
    >>> goal.add_weighted_edges_from([(0,1,n*n),(0,0,n*(n-1)/2), (1, 1, n*(n-1)/2)])  # NOQA
    >>> nx.is_isomorphic(int, goal)
    True
    """
    ret = nx.Graph()
    ret.add_nodes_from(partition.values())

    for node1, node2, datas in graph.edges(data=True):
        edge_weight = datas.get(weight, 1)
        com1 = partition[node1]
        com2 = partition[node2]
        w_prec = ret.get_edge_data(com1, com2, {weight: 0}).get(weight, 1)
        ret.add_edge(com1, com2, **{weight: w_prec + edge_weight})

    return ret


def __renumber(dictionary):
    """Renumber the values of the dictionary from 0 to n
    """
    count = 0
    ret = dictionary.copy()
    new_values = dict([])

    for key in dictionary.keys():
        value = dictionary[key]
        new_value = new_values.get(value, -1)
        if new_value == -1:
            new_values[value] = count
            new_value = count
            count += 1
        ret[key] = new_value

    return ret


def load_binary(data):
    """Load binary graph as used by the cpp implementation of this algorithm
    """
    data = open(data, "rb")

    reader = array.array("I")
    reader.fromfile(data, 1)
    num_nodes = reader.pop()
    reader = array.array("I")
    reader.fromfile(data, num_nodes)
    cum_deg = reader.tolist()
    num_links = reader.pop()
    reader = array.array("I")
    reader.fromfile(data, num_links)
    links = reader.tolist()
    graph = nx.Graph()
    graph.add_nodes_from(range(num_nodes))
    prec_deg = 0

    for index in range(num_nodes):
        last_deg = cum_deg[index]
        neighbors = links[prec_deg:last_deg]
        graph.add_edges_from([(index, int(neigh)) for neigh in neighbors])
        prec_deg = last_deg

    return graph


def __randomly(seq, randomize):
    """ Convert sequence or iterable to an iterable in random order if
    randomize """
    if randomize:
        shuffled = list(seq)
        random.shuffle(shuffled)
        return iter(shuffled)
    return seq


def __one_level(graph, status, weight_key, resolution, randomize):
    """Compute one level of communities
    """
    modified = True
    nb_pass_done = 0
    cur_mod = __modularity(status)
    new_mod = cur_mod

    while modified and nb_pass_done != __PASS_MAX:
        cur_mod = new_mod
        modified = False
        nb_pass_done += 1

        for node in __randomly(graph.nodes(), randomize):
            com_node = status.node2com[node]
            degc_totw = status.gdegrees.get(node, 0.) / (status.total_weight * 2.)  # NOQA
            neigh_communities = __neighcom(node, graph, status, weight_key)
            remove_cost = - resolution * neigh_communities.get(com_node,0) + \
                (status.degrees.get(com_node, 0.) - status.gdegrees.get(node, 0.)) * degc_totw
            __remove(node, com_node,
                     neigh_communities.get(com_node, 0.), status)
            best_com = com_node
            best_increase = 0
            for com, dnc in __randomly(neigh_communities.items(),
                                       randomize):
                incr = remove_cost + resolution * dnc - \
                       status.degrees.get(com, 0.) * degc_totw
                if incr > best_increase:
                    best_increase = incr
                    best_com = com
            __insert(node, best_com,
                     neigh_communities.get(best_com, 0.), status)
            if best_com != com_node:
                modified = True
        new_mod = __modularity(status)
        if new_mod - cur_mod < __MIN:
            break


def __neighcom(node, graph, status, weight_key):
    """
    Compute the communities in the neighborhood of node in the graph given
    with the decomposition node2com
    """
    weights = {}
    for neighbor, datas in graph[node].items():
        if neighbor != node:
            edge_weight = datas.get(weight_key, 1)
            neighborcom = status.node2com[neighbor]
            weights[neighborcom] = weights.get(neighborcom, 0) + edge_weight

    return weights


def __remove(node, com, weight, status):
    """ Remove node from community com and modify status"""
    status.degrees[com] = (status.degrees.get(com, 0.)
                           - status.gdegrees.get(node, 0.))
    status.internals[com] = float(status.internals.get(com, 0.) -
                                  weight - status.loops.get(node, 0.))
    status.node2com[node] = -1


def __insert(node, com, weight, status):
    """ Insert node into community and modify status"""
    status.node2com[node] = com
    status.degrees[com] = (status.degrees.get(com, 0.) +
                           status.gdegrees.get(node, 0.))
    status.internals[com] = float(status.internals.get(com, 0.) +
                                  weight + status.loops.get(node, 0.))


def __modularity(status):
    """
    Fast compute the modularity of the partition of the graph using
    status precomputed
    """
    links = float(status.total_weight)
    result = 0.
    for community in set(status.node2com.values()):
        in_degree = status.internals.get(community, 0.)
        degree = status.degrees.get(community, 0.)
        if links > 0:
            result += in_degree / links - ((degree / (2. * links)) ** 2)
    return result



#!/usr/bin/env python
'''
To construct network:
python run_napa.py -r build -c myconfig_file

Network analysis:
python run_napa.py -r analyze -c myconfig_file
'''

import argparse
from napa.utils.config import *

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--run', required = True,
                        help = ' '.join(['Command to run:',
                        '<build> or <analyze>',
                        'network']))

    parser.add_argument('-c', '--config_file', 
                        required = True,  
                        help = 'Path to configuration file.')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    config = Config(args.config_file, args.run)
    config.print_input_summary()
    
    if 'build' in args.run:
        if 'aln' in config.net_type:
            from napa.analyze.aln_mut_pairs \
                import run_aln_mut_pairs
            run_aln_mut_pairs(config)

        elif 'phylo' in config.net_type: 
            from napa.analyze.phylo_mut_pairs \
                import run_phylo_mut_pairs
            run_phylo_mut_pairs(config)
    
    if 'analy' in args.run: 
        from napa.analyze.net_analysis \
            import run_net_analysis
        run_net_analysis(config)
	      

if __name__ == '__main__':
    main()

	
	      



#! /usr/bin/env python 

# Notes: consider removing ete2/3 dependence
# ete is GPL + only need the PhyloTree class implementation
#
# https://github.com/etetoolkit/ete/blob/master/\
# ete3/phylo/phylotree.py
#
# need to be able to read Newick (*.nwk) files and build a 
# general (nonbinary) phylo tree with attributes on
# nodes and edges 
# FORMAT  DESCRIPTION
# 1       flexible with internal node names

import copy
from scipy.stats import fisher_exact
from ete2 import PhyloTree

from napa.utils.serials import * 
from napa.utils.io import *

from napa.phylo.edge import *
from napa.phylo.tree import get_mrca_subtree

from joblib import Parallel, delayed
import multiprocessing
#import dill

num_cores = multiprocessing.cpu_count()

# Mutation pairs are not allowed to share
# same numerical position
list_mut_pairs = lambda mut_list: \
                 [pair for pair in list_pairs(mut_list) \
                  if get_int_substring(str(pair[0])) != \
                  get_int_substring(str(pair[1]))]

#==============================================================#
class TreeMut(object):
    '''
    Similar to AlnMut class, but counts phylo edges instead of 
    alignment sequences of occurrence.
    '''
    def __init__(self, phylo_edge_pairs = set(), mut_str = None):
        self.mut_str = mut_str
        self.phylo_edge_pairs = set(phylo_edge_pairs)
    
    def add_phylo_edge_pair(self,phylo_edge_pair):
        self.phylo_edge_pairs.add(phylo_edge_pair) 
    
    def add_phylo_edge_pairs(self,phylo_edge_pairs):
        self.phylo_edge_pairs.update(phylo_edge_pairs)

    def __repr__(self):
        return self.mut_str


#==============================================================#
class TreeMutPair(object):
    '''
    Pair of mutations from a single tree.
    Mutations can be on two adjacent edges or
    on occur along the same phylo tree edge.
    '''    
    def __init__(self, phylo_edge_pairs = set(), 
                 mut_pair = None, wt_seq = None,
                 all_edges = set(), 
                 prot_func_transitions = {'':''}):

        # List of pairs of phylo_edge_objects
        if not len(phylo_edge_pairs):
            raise ValueError(\
                ' '.join(['In TreeMutPair:',
                          'Set of Phylo edge pairs', 
                          'must be nonempty!']))
        else:
            self.phylo_edge_pairs = set(phylo_edge_pairs)

        # A unique pair of AA muts
        if mut_pair ==  None or not len(mut_pair):
            raise ValueError('In TreeMutPair: No mutations given!')
        else:
            self.mut_pair = mut_pair
            self.name = '%s_%s' % (mut_pair[0].mut_str, 
                                   mut_pair[1].mut_str)
        if not len(all_edges):
            raise ValueError(' '.join(['In TreeMutPair:',
                                       'No edge pairs having each',
                                       'mutation were given!']))
        else:
            self.all_edges = set(all_edges)
        self.prot_func_transitions = prot_func_transitions

        # Wildtype/ancestral sequence needed to get mutations
        if wt_seq == None:
            raise ValueError('Wild type sequence not provided ' + \
                             str(self))
        else:
            self.wt_seq = wt_seq

    def __repr__(self):
        return '%s\t%s'%(str(self.mut_pair[0]), 
                         str(self.mut_pair[1])) 

    def add_phylo_edge_pair(self, phylo_edge_pair):
        self.phylo_edge_pairs.add(phylo_edge_pair) 
    
    def add_phylo_edge_pairs(self, phylo_edge_pairs):
        self.phylo_edge_pairs.update(phylo_edge_pairs)


    def assign_lineage_distribution(self):
        common_lineage, mut0_lineage, mut1_lineage = [], [], []
        for pep in self.phylo_edge_pairs:
            pe0, pe1 = pep[0], pep[1]                           
            if pe0 == pe1:
                common_lineage  +=  [pe0.parent] + [pe0] + \
                             pe0.followers 
            else:
                common_prec = pe0.get_common_precs(pe1, 
                            'function', self.prot_func_transitions)
                common_follow = list(set([pe0] + pe0.followers) & \
                                    set([pe1] + pe1.followers))
                common_lineage  +=  common_prec + common_follow

                mut0_prec = pe0.get_intermediate_prec(common_prec, 
                            'function', self.prot_func_transitions)
                mut1_prec = pe1.get_intermediate_prec(common_prec, 
                            'function', self.prot_func_transitions)

                mut0_lineage += \
                    list(set(mut0_prec + pe0.followers) \
                         - set(common_prec + common_follow))
                mut1_lineage += \
                    list(set(mut1_prec + pe1.followers) \
                        - set(common_prec + common_follow))
        
        self.common_lineage = set([p for p in common_lineage \
                                   if p and p.muts])  
        self.mut0_lineage = set([p for p in mut0_lineage \
                            if p and p.muts]) - self.common_lineage
        self.mut1_lineage = set([p for p in mut1_lineage \
                            if p and p.muts]) - self.common_lineage
        assoc_lineage = self.common_lineage 
        assoc_lineage |= self.mut0_lineage | self.mut1_lineage
        self.other_lineage = self.all_edges - assoc_lineage
        


    def print_lineages(self):
        stderr_write(["\tCommon Lineage:", 
                      [p.name for p in self.common_lineage]])
        stderr_write(["\tFirst mutation only lineage:", 
                      [p.name for p in self.mut0_lineage]])
        stderr_write(["\tSecond mutations only lineage:", 
                     [p.name for p in self.mut1_lineage]])
        stderr_write(["\tOther mutations lineage:", 
                     [p.name for p in self.other_lineage]])


    def get_contingency_table_lineage_size(self):
        '''
        Contingency table for Fisher's Exact Test.
        Uses overall size of clades where mutations in pair
        occurred. 
        '''
        self.assign_lineage_distribution()
        self.num11 = len(self.common_lineage) # both mutations
        self.num10 = len(self.mut0_lineage) # just the 1st not 2nd
        self.num01 = len(self.mut1_lineage) # just the 2nd not 1st
        self.num00 = len(self.other_lineage) # neither mutation
        self.contingency_table = [[self.num11, self.num01], 
                                  [self.num10, self.num00]]


    def get_contingency_table_edge_counts(self, 
                                          all_phylo_edge_pairs):
        ''' 
        A simpler contingency table based on counts of 
        co-occurrence of mutations in tree edges 
        rather than clades. 
        '''
        if not len(all_phylo_edge_pairs):
            raise ValueError('\nNo pairs of edges in phylogeny'
                             ' have been assigned!!')

        # both mutations
        self.num11 = len(self.phylo_edge_pairs)
        
        # just the 1st not 2nd
        self.num10 = len(self.mut_pair[0].phylo_edge_pairs - \
                         self.phylo_edge_pairs) 

        # just the 2nd not 1st
        self.num01 = len(self.mut_pair[1].phylo_edge_pairs - \
                         self.phylo_edge_pairs)

        # neither mutation
        self.num00 = len(all_phylo_edge_pairs\
                          -self.mut_pair[0].phylo_edge_pairs \
                          -self.mut_pair[1].phylo_edge_pairs)

        self.contingency_table = [[self.num11, self.num01], 
                                  [self.num10, self.num00]]



    def get_raw_count(self):
        '''
        Weight pairs by co-occurrence count only
        '''
        
        self.raw_count = len(self.phylo_edge_pairs)  



    def get_mod_jaccard(self, all_phylo_edge_pairs):
        '''
        Modified Jaccard score for clade / edge count overlap
        '''

        #self.get_contingency_table_lineage_size()
        self.get_contingency_table_edge_counts(all_phylo_edge_pairs)

        [[num11, num10], [num01, num00]] = self.contingency_table

        if (num11 < 2):
            self.mod_jaccard = 0.
            return

        # Penalizes for small co-ocurrence num11
        epsilon = 0.
        #epsilon = float(num00)/(num11 + num10 + num01 + num00)

        self.mod_jaccard = max(0., (num11 - epsilon)) /  \
                           float(num11 + num01 + num10)


    def get_fisher_pval(self):
        '''
        Uses scipy.stats to calculate Fisher's 
        exact p-values based 
        on the contingency table. For association, only need the 
        right-tailed p-value: pval_more.
        '''

        #self.get_contingency_table_lineage_size()
        self.get_contingency_table_edge_counts(all_phylo_edge_pairs)

        self.odds_ratio_less, self.p_val_less = \
            fisher_exact(self.contingency_table, 
                         alternative = 'less')

        self.odds_ratio_more, self.p_val_more = \
            fisher_exact(self.contingency_table, 
                         alternative = 'greater')

#==============================================================#
class TreeMutPairSet(object):

    def __init__(self, aln = None, aln_pos = [], pos_subset = [],
                 wt_seq = None, leaf_fasta_file = '',
                 int_node_fasta_file = '',
                 leaf_seqid_to_prot_func = {'':''}, 
                 int_seqid_to_prot_func = {'':''}, 
                 prot_func_transitions = {}, 
                 sel_prot_func_list = [''], 
                 dist_thresh = -1., 
                 mut_pair_type = 'dir', method = 'mod_jaccard',
                 tree_nwk_file = ''):

        stderr_write(['\nMutation set for tree:\n',
                      tree_nwk_file])

        # Wild-type/ancestral sequence
        if wt_seq == None:
            raise ValueError('Wild-type sequence not provided ' + \
                             'for tree mut pair set based on:\n' \
                             + tree_nwk_file)
        else:
            self.wt_seq = wt_seq

        self.tree_file = tree_nwk_file
        # Add reconstructed internal node sequences to leaf aln
        self.aln = aln.copy()
        self.aln.annot_seqs(annot_key = 'function',
                            seqid_to_annot = \
                            leaf_seqid_to_prot_func)

        self.aln.add_sequences_from_file(int_node_fasta_file)

        # Option to annotate sequences by (reconstructed) function
        # Add reconstructed internal node sequences functions
        self.aln.annot_seqs(annot_key = 'function',
                            seqid_to_annot = \
                            int_seqid_to_prot_func)

        # Names/ids of all tree nodes having selected function
        self.get_sel_func_nodes(annot_key = 'function',
                                annot_list = sel_prot_func_list)


        # Allowed protein function transitions along branches
        self.prot_func_transitions = prot_func_transitions


        # Option to focus on protein region(s)/domain(s) 
        # of interest
        #self.aln.subset_aln_pos(pos_subset = pos_subset)



        # Get the sequence ids of all leaf sequences annotated
        # with the selected function and find most recent 
        # common ancestor
        # This isolates subtree with given function
        self.anc_node, self.sel_func_leaves = \
                    get_mrca_subtree(tree_nwk_file = \
                                     tree_nwk_file,
                                     fasta_file = leaf_fasta_file,
                                     sel_annot_nodes = \
                                     self.sel_func_node_ids)

        # Get muts for seqs corresponding to protein function
        self.get_leaf_func_muts()



        # The folling generates a linegraph object of the 
        # whole tree
        # It is a tree rooted in the edge between anc_node 
        # and its ancestor
        stderr_write(['Building Phylo Edge Tree...'])
        self.anc_edge = PhyloEdge(parent_node = self.anc_node.up, 
                                  child_node = self.anc_node,
                                  aln = self.aln, 
                                  wt_seq = self.wt_seq)
         # Nodes of PhyloEdge are edges are in the original tree
        self.anc_edge.build_edge_tree()

        # Set of all phylogenetic tree edges 
        # When looking at shared edges not lineages
        self.all_phylo_edge_pairs = []

        # Get set of all pairs of mutations occurring along tree
        # default directed mutation pairs
        stderr_write(['Obtaining mutation pairs along tree...'])
        self.mut_pair_type = 'undir' \
                             if 'undir' in mut_pair_type.lower() \
                             else 'dir'
        self.get_mut_pairs(dist_thresh, self.mut_pair_type)

        # Get unique edge pairs
        self.all_phylo_edge_pairs = set(self.all_phylo_edge_pairs)

        for mp in self.mut_pair_to_obj:
            mut_pair = self.mut_pair_to_obj[mp]
            if 'fisher' in method.lower():
                mut_pair.pval_more = 1. # default setting
                mut_pair.get_fisher_pval()
            elif 'jaccard' in method.lower():
                mut_pair.mod_jaccard = 0.
                mut_pair.get_mod_jaccard(self.all_phylo_edge_pairs)
            elif 'raw_count' in method.lower():
                mut_pair.raw_count = 0.
                mut_pair.get_raw_count()
                

    def __repr__(self):
        out_str = ''
        for mp in self.mut_pair_to_weight:
            out_str += '%s\t%.6f\n'%('\t'.join(mp), 
                                     self.mut_pair_to_weight[mp])
        return out_str


    def get_sel_func_nodes(self, annot_key = 'function',
                           annot_list = ['']):
        self.sel_func_node_ids = self.aln.seqids_with_annot(\
                                    annot_key = annot_key,
                                    annot_list = annot_list)
        
    def get_leaf_func_muts(self):
        '''
        Get muts associated with a function that occur 
        in selected tree leaves. 
        This is the subset of all mutations on the 
        tree that is most associated with the function of 
        interest.
        '''
        self.func_muts = []
        for seqid in [leaf.name for leaf in self.sel_func_leaves]:
            self.func_muts += self.wt_seq.get_substitutions(\
                                self.aln.seqid_to_seq[seqid])
        self.func_muts = set(self.func_muts)


    def check_edge_pair(self,start_edge, follow_edge, 
                        dist_thresh):
        if start_edge.is_leaf_edge(): 
            return False
        if not self.check_edge(start_edge, dist_thresh):
            return False
        if not self.check_edge(follow_edge, dist_thresh):
            return False
        if not start_edge.check_dist(follow_edge, dist_thresh):
            return False
        return True

    def check_edge(self, edge, dist_thresh):
        if len(self.prot_func_transitions) and \
           not edge.check_node_seq_annot('function', 
                                self.prot_func_transitions):
            return False

        if len(self.func_muts) and \
           not len(set(edge.muts) & self.func_muts):
            return False

        # Edge is too long -- exceeds distance thresholds
        if not edge.check_dist(edge, dist_thresh):
            return False
            
        return True
        

    def add_phylo_edge_pair(self, phylo_edge_pair, mut_list_pair):
        '''
        Add a phylogeny edge pair on which edges each of the 
        two mutations occurs.
        mut_list_pair: mutation lists for first and second edge
        in phylo_edge_pair
        Also keep track of individual mutation occurrence.
        '''
        mut_set = set(mut_list_pair[0]+ mut_list_pair[1])
        for mut_str in mut_set:
            if mut_str not in self.mut_str_to_obj:
                self.mut_str_to_obj[mut_str] = \
                    TreeMut(phylo_edge_pairs = \
                            set(phylo_edge_pair),
                            mut_str = mut_str)
            else:
                self.mut_str_to_obj[mut_str].\
                    add_phylo_edge_pair(phylo_edge_pair)

        for mp in zip(mut_list_pair[0], mut_list_pair[1]):
            if get_int_substring(mp[0]) == get_int_substring(mp[1]):
                continue
            if mp not in self.mut_pair_to_obj:
                mut_pair = \
                TreeMutPair(\
                wt_seq = self.wt_seq,
                phylo_edge_pairs = set([phylo_edge_pair]),
                mut_pair = [self.mut_str_to_obj[mp[0]],
                            self.mut_str_to_obj[mp[1]]],
                all_edges = self.all_edges)
                self.mut_pair_to_obj[mp] = mut_pair
            else:
                mut_pair = self.mut_pair_to_obj[mp]
                mut_pair.add_phylo_edge_pair(phylo_edge_pair)

        self.all_phylo_edge_pairs.append(phylo_edge_pair)

    def get_mut_pairs(self, dist_thresh, mut_pair_type):
        ''' 
        Extract pairs of mutations occurring along tree edges
        '''
        self.all_edges = [self.anc_edge] + \
                         [fe for fe in self.anc_edge.followers \
                          if fe]

        stderr_write(['\n' + self.tree_file, '\n',
            'Number of tree edges considered:', 
            len(self.all_edges)])

        self.all_edges = [e for e in self.all_edges \
                          if self.check_edge(e, dist_thresh)]

        stderr_write(['\n' + self.tree_file, '\n',
                      'Number of tree edges passing',
                      'distance threshold:', 
                      len(self.all_edges)])

        self.mut_pair_to_obj, self.mut_str_to_obj = {}, {}

        for start_edge in self.all_edges:
            start_edge_sorted_muts = \
                sort_by_digits(list(set(start_edge.muts) & \
                                     self.func_muts))
            if not len(start_edge_sorted_muts): continue
            
            # Make sure mutation pairs occurring on 
            # same edge included in both directed and 
            # undirected networks
            if len(start_edge_sorted_muts) > 1:
                self.add_phylo_edge_pair(\
                    phylo_edge_pair = (start_edge, start_edge), 
                    mut_list_pair = \
                    zip(*list_mut_pairs(start_edge_sorted_muts)))
            
            if 'undir' in mut_pair_type.lower():
                self.add_undir_pairs(start_edge, 
                                     start_edge_sorted_muts, 
                                     dist_thresh)
            else:
                self.add_dir_pairs(start_edge, 
                                   start_edge_sorted_muts, 
                                   dist_thresh)
        
        stderr_write(['\n' + self.tree_file, '\n',       
                      'Number of mutations associated with',
                      'function:', 
                      len(self.mut_str_to_obj.keys())])
        stderr_write(['\n' + self.tree_file, '\n',  
                      'Number of mutation pairs extracted', 
                      len(self.mut_pair_to_obj.keys())])


    def add_undir_pairs(self, start_edge, start_edge_sorted_muts, 
                        dist_thresh):
        for follow_edge in start_edge.followers:
            follow_edge_sorted_muts = \
                sorted(list((set(follow_edge.muts) & \
                             self.func_muts) - \
                            set(start_edge.muts)))

            if self.check_edge_pair(start_edge, follow_edge, 
                                    dist_thresh):
                # For undirected mutation pairs, 
                # mutations are ordered by 
                # residue position
                mut_pairs = [(sm, fm) \
                    for sm in start_edge_sorted_muts \
                    for fm in follow_edge_sorted_muts \
                    if get_int_substring(sm) < get_int_substring(fm)]

                if len(mut_pairs):
                    self.add_phylo_edge_pair(\
                        phylo_edge_pair = (start_edge, follow_edge), 
                        mut_list_pair = zip(*mut_pairs))


            # Reverse order of mutations in cases where follower
            # mutation position is less than start position and distance 
            # from follower to start is less than threshold
            if self.check_edge_pair(follow_edge, start_edge, 
                                    dist_thresh):

                rev_mut_pairs = [(fm, sm) \
                    for sm in start_edge_sorted_muts \
                    for fm in follow_edge_sorted_muts \
                    if get_int_substring(sm) > get_int_substring(fm)]
                
                if len(rev_mut_pairs):
                    self.add_phylo_edge_pair(\
                        phylo_edge_pair = (follow_edge, start_edge), 
                        mut_list_pair = zip(*rev_mut_pairs))
                    

    def add_dir_pairs(self, start_edge, start_edge_sorted_muts, 
                      dist_thresh):
        for follow_edge in start_edge.followers:
            follow_edge_sorted_muts = \
                sorted(list((set(follow_edge.muts) & \
                             self.func_muts) - \
                            set(start_edge.muts)))

            if self.check_edge_pair(start_edge, follow_edge, 
                                    dist_thresh):
                # For undirected mutation pairs, 
                # mutations are ordered by 
                # residue position
                mut_pairs = [(sm, fm) \
                             for sm in start_edge_sorted_muts \
                             for fm in follow_edge_sorted_muts]
                self.add_phylo_edge_pair(\
                    phylo_edge_pair = (start_edge, follow_edge), 
                    mut_list_pair = [start_edge_sorted_muts,
                                     follow_edge_sorted_muts])

#==============================================================#
class TreeEnsembleMutPairs(object):

    def __init__(self, inp):
        self.__dict__.update(vars(inp))
    
        # print the a mutation list for each 
        # sequence in original alignment to 
        # file
        self.print_aln_mut_list()

        self.num_trees = len(flatten(self.tree_files.values()))
        self.mut_pair_to_weight = defaultdict(float)
        self.mut_pair_to_trees = defaultdict(int)

        self.prefix_i = \
        [(prefix, i) for prefix in self.tree_files \
         for i in range(len(self.tree_files[prefix]))]
        

    def __repr__(self):
        out_str = ''
        for mp in self.mut_pair_to_weight: 

            if self.tree_support_thresh > 0:
                if self.mut_pair_to_trees[mp] < \
                   self.tree_support_thresh:
                    continue

            if (('jaccard' in self.method.lower()) or 
                ('raw_count' in self.method.lower())): 
                if self.mut_pair_to_weight[mp] < self.thresh:
                    continue

            if 'fisher' in self.method.lower():
                if self.mut_pair_to_weight[mp] > self.thresh:
                    continue

            out_str += '%s\t%.6f\n' % ('\t'.join(mp), 
                self.mut_pair_to_weight[mp])

        return out_str

    def print_aln_mut_list(self):
        outfile = ''
        if hasattr(self, 'print_seq_muts'):
            if to_bool(self.print_seq_muts):
                if hasattr(self, 'aln'):
                    outfile = \
                    '.'.join(self.aln_fasta_file.split('.')[0:-1])
                    outfile = outfile + '.mutation-list.txt'
                    header = 'sequence_id\tmutations_from_'
                    header += header + self.wt_id.replace(' ','-')
                    if not hasattr(self.aln, 'seqid_to_mut'):
                        self.aln.get_seq_muts(wt_seq = self.wt_seq)

        if len(outfile):            
            stderr_write(['Printing alignment mutations from', 
                          self.wt_id, 'to file', outfile])
            
            write_keyval_dlist(self.aln.seqid_to_mut,
                               outfile, header, ';')


    def assign_mut_pair_weights(self, mut_pair_dicts):
        for dict_pair in mut_pair_dicts:
            mut_pair_to_weight = dict_pair[0]
            mut_pair_to_trees = dict_pair[1]

            self.mut_pair_to_weight = \
            num_dict_update_add(self.mut_pair_to_weight,
                                mut_pair_to_weight)
            
            self.mut_pair_to_trees = \
            num_dict_update_add(self.mut_pair_to_trees,
                                mut_pair_to_trees)
       
    def add_mut_pairs(self, mut_pair_set):
        if 'raw_count' in self.method.lower(): 
            self.add_mut_pairs_raw_count(mut_pair_set)
        elif 'fisher' in self.method.lower():
            self.add_mut_pairs_fisher(mut_pair_set)
        elif 'jaccard' in self.method.lower(): 
            self.add_mut_pairs_mod_jaccard(mut_pair_set)


    def add_mut_pairs_fisher(self, mut_pair_set):
        '''
        to build network, add mutation pairs 
        from a set calculated for a single tree
        all mutations passing the Fisher's exact test 
        with a significant right-tailed p-value are added
        '''
        for mp in mut_pair_set.mut_pair_to_obj:
            mut_pair = \
            mut_pair_set.mut_pair_to_obj[mp]

            if mut_pair.p_val_more <= self.thresh:
                self.mut_pair_to_weight[mp] += \
                        1./self.num_trees
                self.mut_pair_to_trees[mp] += 1
                

    def add_mut_pairs_raw_count(self, mut_pair_set):
        '''
        The weight will be averaged over tree topologies
        '''
        
        for mp in mut_pair_set.mut_pair_to_obj:
            # mut_pair_set for a single tree
            mut_pair = \
                mut_pair_set.mut_pair_to_obj[mp]

            self.mut_pair_to_weight[mp] += \
                mut_pair.raw_count / float(self.num_trees)
            self.mut_pair_to_trees[mp] += 1


    def add_mut_pairs_mod_jaccard(self, mut_pair_set):
        '''
        to build a network, will take the sum of modified 
        Jaccard index and add mutation pairs from a set 
        calculated for a single tree all the mutations 
        with non-zero Jaccard indices.
        The weight will be averaged over tree topologies
        '''
        
        for mp in mut_pair_set.mut_pair_to_obj:
            # mut_pair_set for a single tree
            mut_pair = \
                mut_pair_set.mut_pair_to_obj[mp]

            self.mut_pair_to_weight[mp] += \
                mut_pair.mod_jaccard / float(self.num_trees)
            self.mut_pair_to_trees[mp] += 1
        
    def write_network_to_file(self, file_path):
        '''
        Writes network to file with path file_path
        Columns: source_node  target_node  weight
        '''
        with open(file_path, 'wb') as f:
            f.write(str(self))      
        
               
    def write_table_to_file(self, file_path):
        '''
        Prints a table of all co-occurring mutations.
        Columns: Each mutation in pair,
        weight from phylogeny, 
        number of trees in ensemble passing
        weight threshold.
        '''
        with open(file_path, 'wb') as f:
            f.write('\t'.join(\
                ['Mut1', 'Mut2', 'Weight_' + str(self.method),
                 'Tree_Count']) + '\n')
                    
            for mp in self.mut_pair_to_weight:
                weight = self.mut_pair_to_weight[mp]
                num_trees = self.mut_pair_to_trees[mp]
                f.write('%s\t%.6f\t%d\n' % ('\t'.join(mp), 
                                            weight, num_trees))


    def print_stats(self, aln_pos = [], pos_subset = []):
        stderr_write(['\n\n---------------------------------'])
        stderr_write(['PHYLOGENY-BASED NETWORK PROPERTIES:'])

        length = max(len(aln_pos), self.aln.length)

        stderr_write(['Alignment depth (num. rows)',
                      str(self.aln.depth) + \
                      '\nAlignment length (num. columns):', 
                      length])
        
        stderr_write([len(self.aln.aln_pos), 
                      'positions considered for network.'])

        stderr_write(['Number of ensemble trees considered:',
                      self.num_trees])

        stderr_write([len(self.mut_pair_to_weight), 
                    'Mutation pairs with non-zero',
                    'edge weight.\n'])
        stderr_write(['#######################################'])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_tree_mut_pair_dict(tree_ensemble, prefix, i):
       
    mut_pair_to_weight = defaultdict(float)
    mut_pair_to_trees = defaultdict(int)

    te = tree_ensemble
    tree_file = te.tree_files[prefix][i]

    int_seq_file = te.int_seq_files[prefix][i]

    int_func_file = ''
    if hasattr(te, 'int_func_files'):
        if len(te.int_func_files):
            if len(te.int_func_files[prefix]) > i:
                int_func_file = te.int_func_files[prefix][i]

    int_seqid_to_prot_func = {}        
    if len(int_func_file):
        int_seqid_to_prot_func = \
            parse_keyval_dict(int_func_file)

    mut_pair_set = \
        TreeMutPairSet(\
        aln = te.aln, aln_pos = te.pos_list, 
        pos_subset = te.pos_subset,
        wt_seq = te.wt_seq, 
        leaf_fasta_file = te.aln_fasta_file,
        int_node_fasta_file = int_seq_file,
        leaf_seqid_to_prot_func = \
        te.seqid_to_prot_func, 
        int_seqid_to_prot_func = int_seqid_to_prot_func,
        prot_func_transitions = te.func_transitions,
        sel_prot_func_list = te.sel_prot_func,
        dist_thresh = te.dist_thresh,
        mut_pair_type = te.edge_type, 
        method = te.method, 
        tree_nwk_file = tree_file)

    for mp in mut_pair_set.mut_pair_to_obj:
        mut_pair = mut_pair_set.mut_pair_to_obj[mp]

        if 'raw_count' in te.method.lower(): 
            if mut_pair.raw_count > 0.0:
                mut_pair_to_weight[mp] += \
                mut_pair.raw_count / float(te.num_trees)
                mut_pair_to_trees[mp] += 1

        elif 'fisher' in te.method.lower():
            if mut_pair.p_val_more <= te.thresh:
                mut_pair_to_weight[mp] += \
                                1./te.num_trees
                mut_pair_to_trees[mp] += 1

        elif 'jaccard' in te.method.lower(): 
            if mut_pair.mod_jaccard > 0.0:
                mut_pair_to_weight[mp] += \
                mut_pair.mod_jaccard / float(te.num_trees)
                mut_pair_to_trees[mp] += 1

    return (mut_pair_to_weight, mut_pair_to_trees)

def get_mut_pair_dicts(tree_ensemble):

    mut_pair_dicts = Parallel(n_jobs=num_cores)\
    (delayed(get_tree_mut_pair_dict)\
     (tree_ensemble, p[0], p[1]) \
     for p in tree_ensemble.prefix_i)
    return mut_pair_dicts



#! /usr/bin/env python
import sys
import glob
from collections import defaultdict

from napa.utils.serials import *

'''
Python file objects manipulation and IO.
'''

#=========================================================#
# Standard input/ output/error
#=========================================================#
def stderr_write(inlist):
    sys.stderr.write(\
        ' '.join([str(v) for v in inlist]) + '\n')

#=========================================================#
# Listing files in a directory
#=========================================================#
def list_files(path_before_wildcard = '', 
               path_after_wildcard = '', 
               sort_bool = True):
    if not sort_bool:
        file_list = \
        list(glob.glob(''.join([path_before_wildcard, '*', 
                             path_after_wildcard])))
    else:
        file_list = \
        sorted(list(glob.glob(''.join([path_before_wildcard, 
                            '*', path_after_wildcard]))))
    return file_list

def list_display_files(file_type = '', prefix = '', 
                       suffix = ''):
    stderr_write(['Listing', str(file_type), 'files',
                  prefix, '...', suffix])
    return list_files(prefix, suffix)


#=========================================================#
# General parsing
#=========================================================#
def file_line_iterator(file_name):
    with open(file_name, 'rb') as f:
        for line in f: 
            yield line.strip()

def file_line_list(file_name):
    with open(file_name, 'rb') as f:
        return f.readlines()

def write_table_str(file_name, header, body):
    with open(file_name, 'w') as f:
        f.write(header + body)

#=========================================================#
# Parsing into python serial objects
#=========================================================#
def parse_column(infile, col_num = 0):
    out_list = []
    with open(infile, 'r') as f:
        for line in f:
            recs = line.replace('"','').strip().split()
            if len(recs) > col_num:
                out_list.append(recs[col_num])
    return out_list

def parse_keyval_dict(infile):
    ''' 
    Simple {key:val} dict from first two columns 
    in file 
    '''
    out_dict = {}
    with open(infile, 'rb') as f:
        for line in f:
            recs = [format_str(s) for s in \
                    line.strip().split()]
            if len(recs) > 1:
                out_dict[recs[0]] = recs[1]
    return out_dict

def parse_keyval_dlist(infile):
    ''' 
    List-valued {key:[val0,val1,...]} dict from 
    first two columns in file 
    '''
    out_dict = defaultdict(list)
    with open(infile, 'rb') as f:
        for line in f:
            recs = [format_str(s) for s in \
                    line.strip().split()]
            if len(recs) > 1:
                out_dict[recs[0]].append(recs[1])
    return out_dict


#=========================================================#
# Writing files from python serial objects
#=========================================================#
def write_list_column(out_list = [],
                 outfile = 'list-column.txt', header = ''):
    with open(outfile, 'w') as f:
        if len(header):
            f.write(header + '\n')
        f.write('\n'.join([str(v) for v in out_list]) + '\n')
            

def write_keyval_dict(out_dict = {},
                      outfile = 'dict-keyval.txt', 
                      header = ''):
    with open(outfile, 'w') as f:
        if len(header):
            f.write(header + '\n')
        f.write('\n'.join(['%s\t%s'%(k,v) \
                for k,v in out_dict.iteritems()]) + '\n')

def write_keyval_dlist(out_dict = {},
                       outfile = 'dict-keyvallist.txt', 
                       header = '', val_sep = ';'):

    with open(outfile, 'w') as f:
        if len(header):
            f.write(header + '\n')
        f.write('\n'.join(\
            ['%s\t%s'%(k, 
                       val_sep.join([str(v) for v in l])) \
                       for k, l in out_dict.iteritems()]))
        f.write('\n')



import os

# General parsing/container manipulation
from napa.utils.serials import * 
from napa.utils.io import *

# Configuration and input checking
from napa.utils.config import Config

# Bio sequence manipulation
from napa.seq.parse import *
from napa.seq.bioseq import *

# Network construction from alignment
from napa.mutpair.aln import *


class AlnNetInput(object):
    '''
    Prepare all inputs for alignment-based 
    network reconstruction.
    '''
    def __init__(self, config):
        self.__dict__.update(vars(config))

        #print
        #print '\n'.join(sorted(['%s\t%s'%(k,v) for (k,v) \
        #                        in vars(self).iteritems()]))

        # REQUIRED: Reads wild-type sequence or
        # provides the ID of this sequence in the alignment
        if os.path.isfile(self.wild_type):
            stderr_write(['Reading WT/ancestral sequence',
                          'from file', self.wild_type])
            self.wt_dict = fasta_to_dict(self.wild_type)
            if len(self.wt_dict):
                self.wt_id = self.wt_dict.keys()[0]
                self.wt_seq_str = self.wt_dict[self.wt_id]
        else:
            self.wt_id = self.wild_type
            self.wt_seq_str = ''

        # OPTIONAL: Allows for custom numbering of positions
        # pos_list length must match alignment width
        if len(self.pos_list_file) == 0:
            self.pos_list = []
        else:
            self.pos_list = \
            [int(p) for p in \
             parse_column(self.pos_list_file)]
            stderr_write(['Custom numbering',
                          'specified for',
                          len(self.pos_list), 
                          'alignment positions.'])

        # OPTIONAL: List of protein residue positions 
        # considered for network.
        # Allows removal of residues e.g.
        # from signaling peptide, 
        # highly variable domains/regions.
        if len(self.pos_subset_file) == 0:
            self.pos_subset = []
        else:
            self.pos_subset = \
            [get_int_substring(p) for p in \
             parse_column(self.pos_subset_file)]
            stderr_write(['Custom subset specified.',
                          'It consists of', 
                          len(self.pos_subset), 
                          'alignment positions.'])
            
        # OPTIONAL: Ranges of protein residues to include
        # Can be used instead of pos_subset_file
        if len(self.protein_ranges):
            if len(self.pos_subset):
                stderr_write(['WARNING:Position subset already',
                              'defined in file',
                              self.pos_subset_file,
                              'Protein range list ignored.'])
            else:
                self.pos_subset = set()
                stderr_write([str(self.protein_ranges)])
                for range_str in self.protein_ranges:
                    recs = range_str.strip().split('-')
                    if len(recs) != 2:
                        raise ValueError('Invalid protein' + \
                                'range specified in config:' + \
                                str(recs))
                    begin = get_int_substring(recs[0])
                    end = get_int_substring(recs[1])
                    if not len(self.pos_list):
                        self.pos_subset |= \
                            set(range(begin, end + 1))
                    else:
                        self.pos_subset |= \
                            set([p for p in self.pos_list 
                                 if p in range(begin, end)])
                                           
                self.pos_subset = sorted(list(self.pos_subset))
                stderr_write(['Custom range',
                          'specified. It consists of',
                          len(self.pos_subset), 
                          'alignment positions.'])

        stderr_write(['Position subset consisting of',
                     len(self.pos_subset), 
                     'alignment columns will be',
                      'analyzed.'])
                
        # OPTIONAL: Threshold for including links in network
        # For p-values from Fisher's Exact Test, this is the 
        # maximum p-value; 
        # For Jaccard indices, this is the minimum Jaccard 
        # index weight
        if not hasattr(self, 'thresh'):
            self.thresh = 0.05
            stderr_write(['WARNING: No threshold provided',
                          'for network method', self.method,
                          '\nDefault set to:', self.thresh])
        elif not to_bool(self.thresh) and not \
             self.thresh > 0: 
            self.thresh = 0.05
            stderr_write(['WARNING: No threshold provided',
                          'for network method', self.method,
                          '\nDefault set to:', self.thresh])

        if not hasattr(self, 'min_co_occur'):
            self.min_co_occur = 2
            stderr_write(['WARNING: No minimum co-occurrence',
                          'count provided for network method', 
                          self.method,
                          '\nDefault set to:', self.min_co_occur])

        elif not to_bool(self.min_co_occur):
            self.min_co_occur = 2
            stderr_write(['WARNING: No minimum co-occurrence',
                          'count provided for network method', 
                          self.method,
                          '\nDefault set to:', self.min_co_occur])

        # OPTIONAL: Selection by functional annotation
        # Prot. func. for each seq. in aln with known func. annot.
        if not hasattr(self, 'prot_func_file'):
            self.seqid_to_prot_func = {'function':'default'}
        elif not to_bool(self.prot_func_file):
            self.seqid_to_prot_func = {'function':'default'}
        else:
            self.seqid_to_prot_func = \
                parse_keyval_dict(self.prot_func_file)


        # OPTIONAL: Subset of protein selected functions 
        # considered 
        if not hasattr(self, 'sel_prot_func_file'):
            self.sel_prot_func = ['default']
        elif not to_bool(self.sel_prot_func_file):
            self.sel_prot_func = ['default']
        else:
            self.sel_prot_func = \
            parse_column(self.sel_prot_func_file)

        # Get protein sequence alignment object
        self.get_alignment()
        
        # Get wt sequence id and string
        self.get_wt_seq()




    def print_input_summary(self):
        stderr_write(['Building network from alignment.'])
        stderr_write(['Alignment used:\n', 
                      self.aln_fasta_file])
        

    def get_alignment(self):
        '''
        Alignment from input fasta file
        Do not select sequences by function yet
        (that would remove the WT/reference sequence from set
        '''
        self.aln = \
            BioSeqAln(aln_fasta_file = self.aln_fasta_file,
                      aln_pos = self.pos_list, 
                      pos_subset = self.pos_subset,
                      annot_key = 'function',
                      seqid_to_annot = self.seqid_to_prot_func, 
                      sel_annot_list = self.sel_prot_func)


    def get_wt_seq(self):
        '''
        Obtain WT/ancestral sequence to which all other 
        sequences in the alignment should be compared. 
        Can provide id+sequence, or the id of a sequence 
        already in the alignment.
        If sequence is not in alignment, sequence 
        still needs to match in positions (be aligned) 
        to the multiple sequence alignment.
        No checking or alignment is done.
        '''
        self.wt_id = self.wt_id if len(self.wt_id.strip()) \
                else 'Wild_type'
        
        if len(self.wt_seq_str) == self.aln.length:
            self.wt_seq = \
                BioSeq(seq_id = self.wt_id, 
                       seq_str = self.wt_seq_str,
                       seq_type = 'Protein', 
                       seq_pos_list = self.aln_pos)
        else:
            #make sure wt sequence is trimmed to 
            #the subset of positions (if any)
            #before rejecting due to length mismatch
            if len(self.pos_subset) and \
               self.aln.length == len(self.pos_subset) and \
               len(self.wt_seq_str) >= len(self.pos_subset):
                self.wt_seq = \
                    BioSeq(seq_id = self.wt_id, 
                           seq_str = self.wt_seq_str,
                           seq_type = 'Protein', 
                           seq_pos_list = self.aln.aln_pos, 
                           pos_subset = self.pos_subset)
            
            elif self.wt_id in self.aln.seqid_to_seq:
                if self.wt_seq_str.strip() != '':
                    stderr_write([\
                        'WARNING (AlnMutPairSet):',
                        'Input wild type sequence length',
                        len(self.wt_seq_str), 'does not match', 
                        'alignment length,', self.aln.length,
                        '\nbut sequence id is in alignment.', 
                        '\nReplacing with corresponding',
                        'alignment sequence.'])
                self.wt_seq = self.aln.seqid_to_seq[self.wt_id]
            else:
                raise ValueError(\
                    ' '.join(['ERROR: Invalid  input for',
                              'WT sequence.',
                              '\nPlease enter a valid ID',
                              'already in alignment, '
                              '\n OR a new ID with sequence', 
                              'that aligns to the alignment',
                              '(or aln. column subset chosen).\n',
                              self.wt_id, 'not in\n', 
                              str(sorted(self.aln.seqid_to_seq))]))
            #stderr_write(['WildType positions'] + 
            #             [self.wt_seq.seq_str[i]+
            #              str(self.wt_seq.seq_pos_list[i]) 
            #              for i in range(len(self.wt_seq.seq_str))])



def run_aln_mut_pairs(config):
    '''
    Processes network input information.
    Generates and writes network to text file(s).
    '''
    # Obtains alignment, wt sequence, parameters 
    # for alignment based network reconstruction
    inp = AlnNetInput(config)

    # Calculates edge weights for co-occuring mutation pairs
    aln_mut_pair_set = AlnMutPairSet(inp)



    # Output network in tab-delimited format
    # Source\tTarget\tWeight
    stderr_write(['-----------------------'])
    stderr_write(['\nWriting network to file:\n' + 
                  inp.net_file]) 
    aln_mut_pair_set.write_network_to_file(inp.net_file)

    # outputs additional information about edge weights
    # to a separate table
    if to_bool(inp.output_edge_weight_table):
        stderr_write(['\nWriting network extended network',
                      'table to file:\n' + 
                      inp.net_table_file]) 
        aln_mut_pair_set.write_table_to_file(inp.net_table_file)
 
    # Prints basic stats for reconstructed network
    aln_mut_pair_set.print_stats(aln_pos = inp.pos_list,
                                 pos_subset = inp.pos_subset)



import os
from collections import defaultdict
from copy import deepcopy
from itertools import islice

import math
import random

import networkx as nx
from networkx.utils import weighted_choice
from networkx.algorithms import community

from napa.utils.serials import *
from napa.utils.io import *

#from napa.net.community import *
import napa.net.community_louvain as ml

def set_nx_attribute(g, att_type, att_name, 
                     att_dict):
    ''' 
    Custom version of 
    networkx.classes.function.set_node/edge_attributes
    works with multiple versions of networkx
    '''
    if att_type == 'node':
        for v, att_val in att_dict.iteritems():
            try:
                g.nodes[v][att_name] = att_val
            except KeyError:
                pass
    elif att_type == 'edge':
        for (u, v), att_val in att_dict.iteritems():
            try:
                g[u][v][att_name] = att_val
            except KeyError:
                pass
    return g

def scale_nx_attribute(g, att_type, att_name, 
                       att_dict):
    '''
    Scale a numerical node / edge attribute 
    of a networkx graph
    '''

    if att_type == 'node':
        for v, d in g.nodes.iteritems():
            try:
                g.nodes[v][att_name] *= \
                        float(att_dict[v])
            except (KeyError, ValueError, 
                    AttributeError) as e:
                pass
    elif att_type == 'edge':
        for u, v, d in g.edges(data=True):
            try:
                g[u][v][att_name] *= \
                    float(att_dict[v])
            except (KeyError, ValueError, 
                    AttributeError) as e:
                pass
    return g
        
def most_central_edge(g):
    centrality = nx.edge_betweenness_centrality(g, 
        weight='weight')
    return max(centrality, key=centrality.get)


class MutNet(object):

    '''
    Mutation network analysis methods
    '''

    def __init__(self, net = None, net_file = None, 
                 net_type = None):
        '''
        Networks based on tab-delimited files
        source\ttarget\tweight
        weight: weights from alignment/phylo nets
        net_type: 'directed' or 'undir'ected
        '''
        if net != None:
            self.from_nx_net(net, net_type)
        else:
            self.read_net(net_file, net_type)

        # inverted edges net for some centralities
        self.get_reverse_net()
        
        # inverted weights for centralities 
        #requiring distance
        self.get_inv_w()
        
    #------------------------------------------------------#
    def from_nx_net(self, net, net_type):
        try:
            self.g = deepcopy(net)
            self.directed = nx.is_directed(net)
        except:
            raise Exception('NetworkX format error.')

    #------------------------------------------------------#
    def get_reverse_net(self):
        if self.directed:
            self.rg = nx.reverse(self.g) 
        else:    
            self.rg = deepcopy(self.g)
   
    #------------------------------------------------------#
    def read_net(self, net_file, net_type):
        '''
        Read from file
        Create networkX graph object 
        (directed or undirected)
        '''
        if net_file == None or not os.path.isfile(net_file):
            stderr_write(['WARN: Missing net file. '+ \
                          'Empty network created.'])
            if net_type == 'undir':
                self.g = nx.Graph()
                self.directed = False
            else:
                self.g = nx.DiGraph()
                self.directed = True
            return

        if net_type == 'undir':
            self.g = \
            nx.parse_edgelist(file_line_iterator(net_file), 
                              data=(('weight',float),), 
                              create_using=nx.Graph())
            self.directed = False
        
        else:
            self.g = \
            nx.parse_edgelist(file_line_iterator(net_file), 
                              data=(('weight',float),), 
                              create_using=nx.DiGraph())

            self.directed = True

    #------------------------------------------------------#
    def get_inv_w(self):
        '''
        Weight inversion to convert weights of association to
        distances between linked nodes. 
        The higher the association, the lower the distance.
        Used in all shortest paths and path length methods.
        '''

        if not len(self.g): return
        if not len(nx.get_edge_attributes(self.g, 
                                          'weight').values()):
            return
        max_weight = float(max([d['weight'] for u, v, d in \
                                self.g.edges(data=True)]))
        min_weight = float(min([d['weight'] for u, v, d in \
                                self.g.edges(data=True)]))

        minmax_weight_ratio = min_weight/max_weight
        
        for u, v, d in self.g.edges(data=True):
            self.g[u][v]['weight'] = \
                self.g[u][v]['weight']/max_weight

        for u, v, d in self.g.edges(data=True):
            self.g[u][v]['invWeight'] = \
                minmax_weight_ratio/self.g[u][v]['weight']

        # Reversed graph
        for u, v, d in self.rg.edges(data=True):
            self.rg[u][v]['weight'] = \
                self.rg[u][v]['weight']/max_weight

        for u, v, d in self.rg.edges(data=True):
            self.rg[u][v]['invWeight'] = \
                minmax_weight_ratio/self.rg[u][v]['weight']
            
    #------------------------------------------------------#
    def normalize(self, node_att_dict = {}, norm_factor = 1.):
        norm_factor = 1. if norm_factor == 0. else norm_factor
        d = defaultdict(float)
        for k in node_att_dict:
            d[k] = float(node_att_dict[k]) / norm_factor
        return d

    #------------------------------------------------------#
    def get_centralities(self, path_len_list, 
                         cent_rank_type, 
                         path_len_cent_list, 
                         outfiles):

        for pli, path_len in enumerate(path_len_list):

            cent_list = path_len_cent_list[path_len]

            if path_len == 1:
                self.get_node_centralities(cent_list)

                write_table_str(outfiles[pli], '',
                    self.str_abs_node_centralities(header = True,
                        sorted_cent_names = sorted(cent_list)))

            elif path_len > 1:
                if 'abs' in cent_rank_type:
                    self.get_path_between_path_cent(\
                            path_node_length = path_len)
                    write_table_str(outfiles[pli], '',
                        self.str_path_betw_path_cent())

                elif 'rel' in cent_rank_type:
                    header = 'node.or.path' + \
                     (path_len - 1) * '\t' + \
                     '\t'.join(cent_list) + \
                     '\tnum.net.nodes\n'

                    out_str = \
                    self.get_rel_cent(\
                        cent_list = cent_list, 
                        path_len = path_len)
                    
                    write_table_str(outfiles[pli], 
                                    header, out_str)

    #------------------------------------------------------#
    def get_node_centralities(self, cent_list):
        '''
        Combined standard (local and shortest-path)
        and kpath centrality metrics for 
        single nodes. 

        Normalize by N-1 
        '''

        norm_factor = self.g.number_of_nodes() - 1.

        if 'loc.in.deg' in cent_list:
            self.g = set_nx_attribute(
                self.g, 'node', 'loc.in.deg',
                self.normalize(
                    dict(self.g.in_degree(weight=None)),
                    norm_factor))

        if 'loc.out.deg' in cent_list:
            self.g = set_nx_attribute(
                self.g, 'node','loc.out.deg',
                self.normalize(
                    dict(self.g.out_degree(weight=None)),
                    norm_factor))

        
        if 'loc.deg' in cent_list:
            self.g = set_nx_attribute(
                self.g, 'node','loc.deg',
                self.normalize(
                    dict(self.g.degree(weight=None)),
                    norm_factor))

        if 'loc.strength' in cent_list:
            self.g = set_nx_attribute(
                self.g, 'node', 'loc.strength',
                self.normalize(
                    dict(self.g.degree(weight='weight')),
                    norm_factor))

        if 'loc.in.strength' in cent_list:
            self.g = set_nx_attribute(
                self.g, 'node', 'loc.in.strength',
                self.normalize(
                    dict(self.g.in_degree(weight='weight')),
                    norm_factor))

        if 'loc.out.strength' in cent_list:
            self.g = set_nx_attribute(
                self.g, 'node', 'loc.out.strength',
                self.normalize(
                    dict(self.g.out_degree(weight='weight')),
                    norm_factor))

        #------------------------------------------------------#
        # Note: all (shortest) path-length based centralities 
        # use the inverse weight -- 
        # this is a proxy for distance, rather than assoc.,
        # s.t. stronger association weight = shorter dist.
        if 'glob.close' in cent_list:
            self.g = set_nx_attribute(self.g,'node','glob.close', 
                nx.closeness_centrality(self.g, 
                                        distance = 'invWeight', 
                                        wf_improved = True))

        if 'glob.info' in cent_list:
            self.g = set_nx_attribute(self.g, 'node', 'glob.info', 
                nx.current_flow_closeness_centrality(self.g, 
                    distance = 'weight'))

        if 'glob.eigen' in cent_list:
            try:
                evcent = nx.eigenvector_centrality(self.rg, 
                    max_iter=400, tol=1e-05, weight='weight')
            except nx.exception.NetworkXError:
                evcent = {n:0. for n in self.g.nodes()}

            self.g = set_nx_attribute(self.g, 'node', 
                                      'glob.eigen', evcent)
                                       
        if 'glob.pagerank' in cent_list:
            self.g = set_nx_attribute(self.g, 'node', 
                'glob.pagerank', 
                nx.pagerank(self.rg, weight = 'weight'))


        # betweenness is not distance based in networkx
        if 'glob.betw' in cent_list:
            self.g = set_nx_attribute(self.g, 'node', 
                'glob.betw',
                nx.betweenness_centrality(self.g, 
                    weight = 'weight', normalized = True))

        if 'glob.kpath' in cent_list:
            node_kpath_cents = \
            self.path_kpath_centrality(alpha = 0.)

            for node in self.g.nodes():
                if str(node) in node_kpath_cents:
                    self.g.node[node]['glob.kpath'] = \
                    node_kpath_cents[str(node)]
                else: 
                    self.g.node[node]['glob.kpath'] = 0.

  

    #------------------------------------------------------#
    def str_abs_node_centralities(self, nodes = None,
                                  header = True, 
                                  sorted_cent_names = None):
        '''
        Write centralities for (individual) nodes 
        in the network.
        '''
        
        if sorted_cent_names == None:

            sorted_cent_names = \
            flatten([self.g.node[node].keys() for node in \
                     self.g.nodes()])
            sorted_cent_names = \
            sorted(list(set(sorted_cent_names)))
        
        out_list = []
        if header:
            header = 'node.or.path\t' + \
                     '\t'.join(sorted_cent_names)    
            out_list += [header]

        if nodes != None:
            for node in set(nodes) & set(self.g.nodes()):
                out_list += [str(node) + '\t' + \
                             '\t'.join(['{:.4f}'.format(\
                                    self.g.node[node][cent]) \
                            for cent in sorted_cent_names])]

        else:
            for node in self.g.nodes():
                out_list += \
                [str(node) + '\t' + \
                 '\t'.join(['{:.4f}'.format(\
                    self.g.node[node][cent]) \
                    for cent in sorted_cent_names])]

        if not len(out_list): return ''

        return '\n'.join(out_list) + '\n'

    #------------------------------------------------------#
    def str_node_centralities(self, header = False, 
                              prefix = '', suffix = '',
                              nodes = None, 
                              sorted_cent_names = None):
        '''
        Write centralities for (individual) nodes 
        in the network.
        '''
        
        if sorted_cent_names == None:

            sorted_cent_names = \
            flatten([self.g.node[node].keys() for node in \
                     self.g.nodes()])
            sorted_cent_names = \
            sorted(list(set(sorted_cent_names)))
        
        out_list = []
        if header:
            header = \
            'node.or.path\t' + (prefix + suffix) + '\t' + \
            '\t'.join(sorted_cent_names)+'\tnum.net.nodes'
            
            out_list += [header]

        if nodes != None:
            for node in set(nodes) & set(self.g.nodes()):
                out_list +=  \
                ['%s\t%s\t%s\t' % (prefix, str(node), suffix) + \
                 '\t'.join(['{:.4f}'.format(\
                    self.g.node[node][cent]) \
                    for cent in sorted_cent_names]) + '\t' + \
                 str(self.g.number_of_nodes())]

        else:
            for node in self.g.nodes():
                out_list += \
                ['%s\t%s\t%s\t' % (prefix, str(node), suffix) + \
                 '\t'.join(['{:.4f}'.format(\
                    self.g.node[node][cent]) \
                    for cent in sorted_cent_names]) + '\t' + \
                 str(self.g.number_of_nodes())]

        if not len(out_list): return ''

        return '\n'.join(out_list) + '\n'


    #------------------------------------------------------#
    def get_desc_net_node_cent(self, source = None, 
                               exclude_nodes = [],
                               cent_list = None):

        nodes_in_graph =  nx.descendants(self.g, source)
        nodes_in_graph |= set([source])
        nodes_in_graph -=  set(exclude_nodes)

        if not len(nodes_in_graph): 
            desc_net = MutNet()
        
        desc_net = \
        MutNet(net = self.g.subgraph(nodes_in_graph))
        
        desc_net.get_node_centralities(cent_list = cent_list)
        return desc_net

    #------------------------------------------------------#
    def get_rel_cent(self, cent_list = None, prev_nodes = [], 
                     nodes = None, path_len = 3):
        
        if len(prev_nodes) >= path_len: return ''
        
        if nodes != None and not len(nodes): nodes = None

        
        prefix = '\t'.join(prev_nodes)
        suffix = (path_len - len(prev_nodes) - 1)*'\t----'
        
        self.get_node_centralities(cent_list = cent_list)
        out_str = \
        self.str_node_centralities(prefix = prefix,
                                   suffix = suffix,
                                   nodes = nodes)

        for node in self.g:

            if node in prev_nodes: continue
            desc_net = \
            self.get_desc_net_node_cent(source = node,
                            exclude_nodes = prev_nodes,
                                cent_list = cent_list)

            if desc_net.g.size() <= 3: continue

            out_str += \
            desc_net.get_rel_cent(cent_list = cent_list, 
                prev_nodes = prev_nodes + [node],
                nodes = list(set(self.g.neighbors(node)) - \
                             set(prev_nodes)),
                path_len = path_len)

        return out_str

    #------------------------------------------------------#
    def get_path_between_path_cent(self, path_node_length = 2,
                    cent_list = ['glob.betw', 'glob.kpath']):
        '''
        Get shortest and k-path betweenness centralities for 
        all paths consisting of path_node_length number of nodes.
        Here path_len refers to number of nodes rather than 
        edges in path.
        '''
        self.path_cent = {}
        
        if 'glob.betw' in cent_list:
            self.path_shortest_path_b_cent(\
                                path_len = path_node_length)

        if 'glob.kpath' in cent_list:            
            self.path_cent['betw_kpath'] = \
                deepcopy(self.path_kpath_centrality(alpha = 0., 
                        path_len = path_node_length))

    #------------------------------------------------------#
    def str_path_betw_path_cent(self):

        sorted_cent_names = sorted(self.path_cent.keys())
        out_str =  '\t'.join(['Path'] + \
            [cent for cent in sorted_cent_names]) + '\n'
            
        path_set = set()
        for path_dict in self.path_cent.values():
            path_set |= set(path_dict.keys())

        for path in path_set:
            out_str += path + '\t' + \
                '\t'.join(['%.9f'%(self.path_cent[cent][path]) \
                           if path in self.path_cent[cent] \
                           else '%.9f'%(0.) for cent in \
                           sorted_cent_names]) + '\n'
        return out_str

                                            
    #------------------------------------------------------#
    def all_sps(self, source, target):
        '''
        No error all_shortest_paths between source and target
        '''
        try:
            return \
            [p for p in nx.all_shortest_paths(self.g, 
                        source, target, weight = 'invWeight')]

        except nx.NetworkXNoPath:
            return []

    #------------------------------------------------------#
    def path_shortest_path_b_cent(self, path_len = 2):
        '''
        Extension of the single-node shortest path (sp)
        betweenness to a set of contiguous nodes 
        in the network (can be further optimized)
        '''
        num_all_paths = 0
        self.path_cent['betw_shortest_path'] = defaultdict(float)

        for source in self.g.nodes():

            for target in self.g.nodes():
                if source == target: 
                    continue #better iteration strategy?

                st_paths = self.all_sps(source, target)    
                num_all_paths += len(st_paths)
                for st_path in st_paths:
                    sub_paths = \
                    [wn for wn in window(st_path[1:-1], 
                                         n = path_len)]

                    for sub_path in sub_paths:
                        sub_path_str_list = \
                        [str(node) for node in sub_path]

                        if path_len == 1:
                            self.path_cent['betw_shortest_path'][\
                                '_'.join(sub_path_str_list)] \
                                += 2
                            continue

                        if not self.directed:
                            self.path_cent['betw_shortest_path'][\
                            '_'.join(\
                                reversed(sub_path_str_list))] \
                                += 1

                            self.path_cent['betw_shortest_path'][\
                            '_'.join(sub_path_str_list)] += 1

                        else:
                            self.path_cent['betw_shortest_path'][\
                                '_'.join(sub_path_str_list)] \
                                += 2

        for path in self.path_cent['betw_shortest_path']:
            self.path_cent['betw_shortest_path'][path] *= \
                                    (0.5 / num_all_paths)


    #------------------------------------------------------#
    def path_kpath_centrality(self, k = None, alpha = 0.2, 
        weight = 'invWeight', seed = 123456, path_len = 1):
        '''
        Adapted from Alakahoon et al. and published code
        by extending from single node to path centrality
        '''

        kpath_path_cent = defaultdict(float)

        if self.g.is_multigraph():
            raise nx.NetworkXError("Not implemented" + \
                                   "for multigraphs.")

        n = self.g.number_of_nodes()
        ne = self.g.number_of_edges()

        if n <= path_len + 4 or ne < path_len + 4:
            # network is too small for calculating
            # path centr. of this length
            return {}
            
        if k is None: 
            k = int(math.log(n + ne))

        if k < path_len + 4: k = min(path_len + 4, n)
            
        T = 2. * k**2 * n**(1-2*alpha) * math.log(n)
        
        random.seed(seed)

        for i in range(int(T+1)):
            st_path = []
            
            # choose source node
            s = random.choice(list(self.g.nodes))

            # choose a random path length
            l = random.randint(path_len + 2, k) 
            st_path.append(s)

             # fill out a path of length l
            for j in range(l):
                # invert distance again - 
                # get association
                nbrs = {nbr: 1./d.get(weight, 1.0) \
                        for nbr,d in self.g[s].items() \
                        if nbr not in st_path} 

                if not nbrs: break

                v = weighted_choice(nbrs)
                st_path.append(v)

                # set the current path source 
                # (current node) to v
                s = v 

            if len(st_path) < path_len + 4: continue

            # sub_paths exclude end-points
            sub_paths = [wn for wn in window(st_path[1:-1], 
                                             n = path_len)]

            for sub_path in sub_paths:
                sub_path_str_list = \
                [str(node) for node in sub_path] 

                if path_len == 1:
                    kpath_path_cent[\
                        '_'.join(sub_path_str_list)] += 2.
                    continue

                if not self.directed:
                    kpath_path_cent['_'.join(\
                        reversed(sub_path_str_list))] += 1.

                    kpath_path_cent[\
                        '_'.join(sub_path_str_list)] += 1.

                else:
                    kpath_path_cent[\
                        '_'.join(sub_path_str_list)] += 2.


        for path in kpath_path_cent:
            kpath_path_cent[path] *= (0.5 / T)

        
        return kpath_path_cent

    #------------------------------------------------------#
    #----------------------COMMUNITIES---------------------# 
    def get_node_clusters(self, node_clust_file, 
                          comm_types = ['MLV']):
        #comm_types = ['GNB', 'ALP', 'MLV']
        self.communities = {}

        # Community detection algorithms not defined on
        # directed graphs, so need to:
        # deep copy with automatic conversion to undirected
        self.ugc = self.g.to_undirected() 


        if 'GNB' in comm_types:
            # Get Girwan-Newman edge betw. communities
            self.get_gn_communities()
        
        if 'ALP' in comm_types: 
            #Get asynchronous label propagation communities
            self.get_alp_communities()

        if 'MLV' in comm_types:
            # Louvain multilevel communities
            self.get_multilevel_communities()

        for comm_name in self.communities:
            set_nx_attribute(self.g, 'node', comm_name,
                             self.communities[comm_name])

        # Write desired community partitions to file
        self.write_communities(self.communities.keys(),
                               node_clust_file)

    #------------------------------------------------------# 
    def get_gn_communities(self):

        comm_generator = community.girvan_newman(
            self.ugc, most_central_edge)
        comm_part = next(comm_generator) # top level
        node_comm = {}
        for comm_i, node_list in enumerate(comm_part):
            for node in node_list:
                node_comm[node] = comm_i

        comm_mod = community.quality.modularity(
            G=self.ugc, communities = comm_part,
            weight = 'weight')

        self.communities['comm.GNB_%.2f'%comm_mod] = \
                                        node_comm
            
    #------------------------------------------------------# 
    def get_alp_communities(self): 
        comm_iterator = community.asyn_lpa_communities(
            self.ugc, weight='weight')
        comm_part = list(comm_iterator)
        node_comm = {}
        for comm_i, node_set in enumerate(comm_part):
            for node in node_set:
                node_comm[node] = comm_i

        comm_mod = community.quality.modularity(
            G=self.ugc, communities = comm_part, 
            weight='weight')

        self.communities['comm.ALP_%.2f'%comm_mod] = \
                                            node_comm

    #------------------------------------------------------# 
    def get_multilevel_communities(self): 
        '''
        Get hierarchical multilevel communities.
        '''

        partitions, modularities = [], []
        resolutions = [0.5, 1., 2., 4., 8.]
        for r in resolutions:
            best_partition = ml.best_partition(
                graph = self.ugc, weight = 'weight', 
                resolution = r)
            partitions.append(best_partition)
            modularities.append(ml.modularity(
                best_partition, self.ugc, 'weight'))
 
        max_mod, max_idx = 0., 1
        for im, mod in enumerate(modularities):
            if mod > max_mod:
                max_mod = mod
                max_idx = im
                
        num_partitions = len(set(partitions[max_idx].values()))
        comm_part = [set() for _ in range(num_partitions)]
        for (node, comm) in partitions[max_idx].iteritems():
            comm_part[comm].add(node)

        comm_mod = community.quality.modularity(G=self.ugc, 
                communities = comm_part, weight = 'weight')

        self.communities['comm.MLV_%.2f'%comm_mod] = \
                                    partitions[max_idx]

        
    #------------------------------------------------------#
    def write_communities(self, comm_names, node_clust_file):
        '''
        Writes node attributes calculated with each method
        '''
        stderr_write(['Writing community partitions:\n',
                      ', '.join(comm_names)])

        with open(node_clust_file, 'wb') as f:
            f.write('Node\t'+ '\t'.join(comm_names) + '\n')
            
            for node in self.g.nodes():
                f.write('\t'.join([str(node)] + \
                        [str(self.g.node[node][cn]) \
                         for cn in comm_names \
                         if cn in self.g.node[node]]) \
                        + '\n')




#! /usr/bin/env python
import string
import re
from collections import defaultdict
from itertools import islice
'''
Python serial objects manipulation.
'''

#=========================================================#
# List/tuple manipulation
#=========================================================#
flatten = lambda x: \
    [subitem for item in x for subitem in item]

list_pairs = lambda x: \
    [(x[i1],x[i2]) \
     for i1 in range(len(x)-1) \
     for i2 in range(i1+1,len(x))]

sort_list_by_other = lambda x,y: \
    [x for (y,x) in sorted(zip(y,x), 
                           key=lambda pair: pair[0])]

def window(seq, n=2):
    '''
    Returns a sliding window (of width n) over data 
    from the iterable:
    s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ... 
    '''
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result    
    for elem in it:
        result = result[1:] + (elem,)
        yield result

#=========================================================#
# Dictionaries
#=========================================================#
def num_dict_update_add(d0, d1):
    '''
    Update dictionary with numerical values from  another
    '''
    for k in d1:
        d0[k] += d1[k]

    return d0


many_to_one = lambda dl: \
              {v:k for k,l in dl.iteritems() for v in l}

        

#=========================================================#
# String manipulation
#=========================================================#
# Remove punctuation except for underscore
format_str = lambda s: \
    s.translate(string.maketrans("", ""), 
                string.punctuation.replace('_', ''))
#---------------------------------------------------------#
# Integer substring
get_int_substring = lambda s: int(re.sub('\D','',s))

# Get number from str

def get_int_from_str(s, default):
    try:
        return int(s)
    except ValueError:
        return default

def get_float_from_str(s, default):
    try:
        return float(s)
    except ValueError:
        return default

#---------------------------------------------------------#
# Lists of strings (sort by integer substring)
sort_by_digits = \
lambda x: sorted(x, key = lambda s: get_int_substring(s))

#=========================================================#
# Strings or numbers to Boolean 
#=========================================================#
def to_bool(v):
    '''
    Also interprets no/false strings as False
    '''
    if isinstance(v, str):
        if v.lower() in [ '', '0', 'false', 'f', 'n', 'no']:
            return False

    return bool(v)





#!/usr/bin/env python
from collections import defaultdict

from napa.utils.config import Config
from napa.utils.serials import *
from napa.utils.io import *

from napa.net.net import *

class NetAnalysisInput(object):
    '''
    Prepare all inputs for phylogeny-based 
    network reconstruction.
    Builds on alignment inputs as provided for 
    alingment based network and adds 
    phylogeny specific parameters
    '''
    def __init__(self, config):
        self.__dict__.update(vars(config))
            
        if self.calculate_centralities:
            self.get_path_len_list()
            self.cent_list = defaultdict(list)
            for pl in self.path_len:
                self.get_cent_list(pl)
        
            self.cent_files = \
                [self.net_file.replace('.net.txt', 
                    '.'.join(['.cent_' + self.cent_type,
                              'path_len' + str(pl),
                              self.cent_rank_type[0:3],
                              'txt'])) \
                 for pl in self.path_len]

            stderr_write(['\nCentralities will be calculated',
                          'and printed to:\n\t' + \
                          '\n\t'.join(self.cent_files)])

        if self.cluster_nodes:
        
            self.node_clust_file = \
                self.net_file.replace('.net.txt', 
                                      '.communities.txt')

            stderr_write(['\nNode communities will be',
                          'printed to:\n',
                          self.node_clust_file])
            
    #--------------------------------------------------------#
    def get_path_len_list(self):

        if self.path_len == None:
            self.path_len = [1]
            stderr_write(['\nPath length not provided',
                'for path or rel. path centr.',
                '\nSetting it to default 1.'])
            
        if 'rel' in self.cent_rank_type.lower():
            path_len = [pl for pl in self.path_len if pl > 1]
            if not path_len or path_len != self.path_len:
                if not len(path_len):
                    self.path_len = [2]
                else:
                    self.path_len = path_len
                
                stderr_write(['\nWARNING: Path lengths',
                    '< 2 not defined for relative centralities.',
                    'Calculating centralities for path lengths:'] +\
                    self.path_len)

    #---------------------------------------------------------#
    def get_cent_list(self, path_len):
        '''
        Get all centralities that can be calculated based 
        on applicability to directed or undirected graphs.
        '''
       

        if 'loc' in self.cent_type or \
           'both' in self.cent_type:
            if self.edge_type.startswith('dir'):
                # Directed local centralities
                self.cent_list[path_len] +=  \
                ['loc.in.deg', 'loc.out.deg', 
                 'loc.in.strength', 
                 'loc.out.strength']
            else:
                # Undirected local centralities
                self.cent_list[path_len] += \
                ['loc.deg', 'loc.strength']

        if 'glob' in self.cent_type or \
           'both' in self.cent_type:
            self.cent_list[path_len] +=  \
                ['glob.close', 'glob.eigen', 
                 'glob.pagerank', 
                 'glob.betw', 'glob.kpath']

        if path_len > 1 and 'abs' in \
           self.cent_rank_type:

            self.cent_list[path_len] =  \
            ['glob.betw', 'glob.kpath']
            
                
        stderr_write(['\nWill calculate', 
                      self.cent_rank_type,
                      'centralities, for', 
                      self.cent_type,
                      'centrality types:\n'] + \
                     self.cent_list[path_len])

        if not len(self.cent_list[path_len]):
            stderr_write(['\nInvalid centrality',
                          'type provided',
                          'calculating undirected degree',
                          'centrality only.'])
            self.cent_list[path_len] = ['loc.deg']
            

#=========================================================#
def run_net_analysis(config):
    '''
    Processes network input information.
    Generates and writes network to text file(s).
    '''
    inp = NetAnalysisInput(config)
    mn = MutNet(net_file = inp.net_file, 
                net_type = inp.edge_type)

    if to_bool(inp.calculate_centralities):
        mn.get_centralities(\
            path_len_list = inp.path_len,
            cent_rank_type = inp.cent_rank_type,
            path_len_cent_list = inp.cent_list,
            outfiles = inp.cent_files)

    if to_bool(inp.cluster_nodes):
        mn.get_node_clusters(inp.node_clust_file, 
                             inp.partition_types)
    



#! /usr/bin/env python 

from collections import defaultdict
from scipy import stats

from napa.utils.serials import * 
from napa.utils.io import *

from napa.seq.bioseq import *

def add_seq_id_to_set(seq_id, curr_set):
    '''
    expands current set with a (str of) single 
    sequence id or or a (listlike of) multiple
    ids
    '''
    if not isinstance(seq_id, basestring):
        curr_set.update(set(seq_id))
    else:
        curr_set.add(seq_id)
    return curr_set

# Mutation pairs are not allowed to share
# same numerical position
list_mut_pairs = lambda mut_list: \
                 [pair for pair in list_pairs(mut_list) \
                  if get_int_substring(str(pair[0])) != \
                  get_int_substring(str(pair[1]))]

class AlnMut(object):

    def __init__(self, seqs=set(), mut_str = None):
        self.mut_str = mut_str
        self.seqs = add_seq_id_to_set(seqs, set())
    
    def add_seq(self,seq):
        self.seqs = add_seq_id_to_set(seq, self.seqs)

    def __repr__(self):
        return self.mut_str


class AlnMutPair(object):
    '''
    Pair of mutations from seq alignment.
    '''
    def __init__(self, seqs = set(), mut_pair = None):

        # Unique set of seqs containing mutation
        if not len(seqs):
            raise ValueError(\
                "No sequences or empty sequence IDs "+ \
                    "associated with mut pair.")
        else:
            if not isinstance(seqs, basestring):
                self.seqs = set(seqs)
            else:
                self.seqs = set([seqs])

        if not mut_pair:
            raise ValueError('No mutations given ' + \
                             'for mutation pair')
        else:
            # Mutations in mut_pair should be AlnMut instances
            self.mut_pair = mut_pair 

    def __repr__(self):
        return '%s\t%s'%(self.mut_pair[0], self.mut_pair[1])

    def add_seq(self,seq):
        self.seqs = add_seq_id_to_set(seq, self.seqs)

    def get_contingency_table(self, all_seqs):
        ''' 
        Contingency table of occurrence counts of
        mutations, occurring individually and together. 
        Occurrence count = number of seqs containing mut.
        '''

        occur01 = len(self.seqs) #occurrence count of both muts

        # Number of seqs containing first or second mut, not both 
        occur0 = len(self.mut_pair[0].seqs - self.seqs) 
        occur1 = len(self.mut_pair[1].seqs - self.seqs)

        # Number of seqs containing neither mut
        occur_other = len(all_seqs \
                        - self.mut_pair[0].seqs \
                        - self.mut_pair[1].seqs)

        self.contingency_table = [[occur01, occur0], 
                                  [occur1, occur_other]]
        self.num_seqs = len(all_seqs)
        

    def get_raw_count(self, min_co_occur = 2):
        '''
        Weight pairs by co-occurrence count only
        '''
        [[num11, num10], 
         [num01, num00]] = self.contingency_table

        if num11 < min_co_occur:
            self.weight = 0.
            return
        
        self.weight = num11

    def get_jaccard(self, min_co_occur = 2):
        ''' 
        OBSOLETE 
        Jaccard weight with 
        1. minimum requirement for coocurrence count
        2. if no minimum is set, introduce correction 
        for pairs of muts occurring only once
        and only together in one seq.
        '''
        [[num11, num10], 
         [num01, num00]] = self.contingency_table

        if num11 < min_co_occur:
            self.weight = 0.
            return

        # Penalizes for small co-ocurrence num11
        epsilon = 1./self.num_seqs

        self.weight = max(0.,
            (num11 - epsilon) / float(num11 + num01 + num10))


    def get_mod_jaccard(self, min_co_occur = 2):
        '''
        Modified Jaccard score based on sequence occurrence counts
        corrected for low mutation co-occurrence counts
        '''
        [[num11, num10], [num01, num00]] = self.contingency_table
        
        if (num11 < min_co_occur):
            self.weight = 0.
            return

        # Penalizes for small co-ocurrence num11
        epsilon = 0. # no penalty
        #epsilon = float(num00)/(num11 + num10 + num01)
        
        self.weight = max(0.,(num11 - epsilon)) /   \
                      float(num11 + num01 + num10)


    def get_fisher_pval_weight(self, min_co_occur = 2):
        ''' 
        Unlike Jaccard, the smaller the p-value, 
        the greater the association '''

        [[num11, num10], 
         [num01, num00]] = self.contingency_table
        
        if num11 < min_co_occur:
            self.odds_ratio_more = float(num11 * num00)
            self.odds_ratio_more /= num10 * num01
            self.odds_ratio_less = float(num10 * num01)
            self.odds_ratio_less /= num11 * num00
            self.pval_less = 1.
            self.pval_more = 1.
            return

        self.odds_ratio, self.pval_less = \
        stats.fisher_exact(self.contingency_table, 
                           alternative = 'less')

        self.odds_ratio, self.pval_more = \
        stats.fisher_exact(self.contingency_table, 
                           alternative = 'greater')

        self.weight = self.pval_more

class AlnMutPairSet(object):
    '''
    Set of mut pairs and associated metrics from an MSA.
    '''
    def __init__(self, inp):
        self.__dict__.update(vars(inp))

        # Pick out sequences with selected function of interest
        # (Alignment may consist of sequences with different 
        # annotations)
        self.aln.subset_annot_seq_dict(\
            sel_annot_key = 'function', 
            sel_annot_list = self.sel_prot_func)

        self.aln.get_seq_muts(self.wt_seq)

        # Print mutations from WT for each sequence in aln
        self.print_aln_mut_list()

        # Get all possible mut pairs in alignment, 
        #given selected function
        self.get_aln_mut_pairs()


    def __repr__(self):
        out_str = ''
        for mp in self.mut_pair_to_obj:
            
            mut_pair = self.mut_pair_to_obj[mp]
    
            if (('jaccard' in self.method.lower()) or 
                ('raw_count' in self.method.lower())): 
                if mut_pair.weight < self.thresh:
                    continue

                out_str += \
                '%s\t%.6f\n' % (mut_pair, 
                                mut_pair.weight)

            elif 'fisher' in self.method.lower(): 
                if mut_pair.weight > self.thresh:
                    continue

                out_str += \
                '%s\t%.6f\n' % (mut_pair, 
                                mut_pair.pval_more)

        return out_str


    def get_aln_mut_pairs(self):
        ''' 
        Extract pairs of mutations occurring 
        in at least one seq.
        '''

        self.mut_str_to_obj = {} 
        self.mut_pair_to_obj = {}
        stderr_write(['Extracting pairs of',
                      'mutations from', 
                      len(self.aln.seqid_to_mut), 
                      'sequences.'])

        for seqid in self.aln.seqid_to_mut:
            for m in self.aln.seqid_to_mut[seqid]:
                if m not in self.mut_str_to_obj:
                    self.mut_str_to_obj[m] = \
                        AlnMut(seqs = seqid, mut_str = m)
                else:
                    self.mut_str_to_obj[m].add_seq(seqid)

            for mp in \
                list_mut_pairs(self.aln.seqid_to_mut[seqid]):
                if mp not in self.mut_pair_to_obj:
                    self.mut_pair_to_obj[mp] = \
                    AlnMutPair(seqs = seqid, 
                               mut_pair = \
                            (self.mut_str_to_obj[mp[0]],
                             self.mut_str_to_obj[mp[1]]))
                else:
                    self.mut_pair_to_obj[mp].add_seq(seqid)

        # get contingency tables and network weights 
        # for pairs of co-occurring mutations
        all_seqs = set(self.aln.seqid_to_mut.keys())

        if not len(self.mut_pair_to_obj):
            stderr_write(['WARNING:', 'No mutation pairs found!',
                          '\n\tEmpty network will be produced'])
        else:
            stderr_write(['Calculating weights for', 
                          len(self.mut_pair_to_obj),
                          'mutation pairs.'])

        for mp in self.mut_pair_to_obj:           
            mut_pair = self.mut_pair_to_obj[mp]
            
            mut_pair.get_contingency_table(all_seqs = \
                                           all_seqs)
            
            if 'raw_count' in self.method.lower():
                mut_pair.get_raw_count(self.min_co_occur)

            if 'mod_jaccard' in self.method.lower():
                mut_pair.get_mod_jaccard(self.min_co_occur)

            elif 'jaccard' in self.method.lower():
                mut_pair.get_jaccard(self.min_co_occur)
            
            elif 'fisher' in self.method.lower():
                mut_pair.get_fisher_pval_weight(\
                                self.min_co_occur)
            
    def print_aln_mut_list(self):
        outfile = ''
        if hasattr(self, 'print_seq_muts'):
            if to_bool(self.print_seq_muts):
                if hasattr(self, 'aln'):
                    outfile = \
                    '.'.join(self.aln_fasta_file.split('.')[0:-1])
                    outfile = outfile + '.mutation-list.txt'
                    header = 'sequence_id\tmutations_from_'
                    header += header + self.wt_id.replace(' ','-')

        if len(outfile):            
            stderr_write(['Printing alignment mutations from', 
                          self.wt_id, 'to file', outfile])
            
            write_keyval_dlist(self.aln.seqid_to_mut,
                               outfile, header, ';')

    def write_network_to_file(self, file_path):
        '''
        Writes network to file with path file_path
        Columns: source_node  target_node  weight
        '''
        with open(file_path, 'wb') as f:
            f.write(str(self))      
                    

    def write_table_to_file(self, file_path):
        '''
        Prints a table of all co-occurring mutations.
        Columns: Each mutation in pair, number of co-occurrences,
        contingency table (for Fisher's exact) and Jaccard Index.
        '''
        with open(file_path, 'wb') as f:
            f.write('\t'.join(\
                ['Mut1', 'Mut2', 'Weight_' + str(self.method),
                 'Contingency_Table', 'Co-occur_Count', 
                 'Sequence_Sources']) + '\n')
                    
            for mp in self.mut_pair_to_obj:
                mut_pair = self.mut_pair_to_obj[mp]
                
                f.write('%s\t%.6f\t%s\t%d\t%s\n' % \
                        (mut_pair, mut_pair.weight, 
                         mut_pair.contingency_table,
                         len(mut_pair.seqs),
                         ';'.join(sorted(list(mut_pair.seqs)))))
     

    def print_stats(self, aln_pos = [], pos_subset = []):
        stderr_write(['\n\n------------------------------'])
        stderr_write(['ALIGNMENT NETWORK PROPERTIES:'])

        length = max(len(aln_pos), self.aln.length)

        stderr_write(['Alignment depth (num. rows)',
                      str(self.aln.depth) + \
                      '\nAlignment length (num. columns):', 
                      length])

        stderr_write([len(self.aln.aln_pos), 
                      "positions considered for network."])

        stderr_write([len([self.mut_pair_to_obj[mp].weight
                           for mp in self.mut_pair_to_obj 
                           if (self.mut_pair_to_obj[mp].weight 
                               >= self.thresh)]),
                      "mutation pairs in network.\n"])




from collections import defaultdict

from napa.utils.io import *
from napa.utils.serials import *

from napa.mutpair.phylo import * 
from napa.analyze.aln_mut_pairs import *

class PhyloNetInput(AlnNetInput):
    '''
    Prepare all inputs for phylogeny-based 
    network reconstruction.
    Builds on alignment inputs as provided for 
    alingment based network and adds 
    phylogeny specific parameters
    '''
    def __init__(self, config):
        # (Leaf sequence alignment inputs)
        super(PhyloNetInput, self).__init__(config)

        # Phylogeny specific inputs
        self.func_transitions = {}
        not_specified = True
        if bool(self.func_transitions_file):
            if len(self.func_transitions_file):
                    self.func_transitions = \
                        parse_keyval_dict(self.func_transitions_file)
                    stderr_write(['Using', len(self.func_transitions),
                                  'functional transitions on tree.'])
                    not_specified = False
        
        if not_specified:
            stderr_write(['Functional transitions between',
                          'internal node functions when',
                          'traversing branches',
                          'not constrained.'])
        self.get_tree_files()

    def get_tree_files(self):
        
        self.tree_files =  defaultdict(list)
        self.int_seq_files = defaultdict(list)
        self.int_func_files = defaultdict(list)

        stderr_write(['Retrieving tree file(s).'])
        for prefix in self.tree_file_prefix_list:
            if bool(prefix):
                prefix = os.path.join(self.working_dir,
                                      self.data_dir, prefix)
            else:
                prefix = os.path.join(self.working_dir,
                                      self.data_dir)+'/'

            self.tree_files[prefix] = \
                list_display_files('Nwk trees with internal '+ \
                                   'nodes', prefix, 
                                self.tree_nwk_file_suffix)

            self.int_seq_files[prefix] = list_display_files(\
                'Internal node sequences', prefix, 
                self.tree_internal_node_seq_suffix)

            if hasattr(self,'tree_internal_node_state_suffix'):
                if self.tree_internal_node_state_suffix != None:
                    self.int_func_files[prefix] = \
                        list_display_files('Reconstructed ' + \
                            'functions on internal nodes', 
                                           prefix, 
                            self.tree_internal_node_state_suffix)

            if len(self.int_seq_files[prefix]) != \
               len(self.tree_files[prefix]):
                raise ValueError(\
                    ' '.join(['PhyloNet Build Input:',
                              'Check that each Newick tree file',
                              'has a corresponding',
                              'internal states file!']))

            if len(self.int_func_files[prefix]):
                if len(self.int_func_files[prefix]) != \
                   len(self.tree_files[prefix]):
                    raise ValueError(\
                        ' '.join(['PhyloNet Build Input:',
                                  'Check that each Newick tree',
                                  'file has a corresponding', 
                                  'internal states file!']))


def run_phylo_mut_pairs(config):
    '''
    Processes network input information.
    Generates and writes network to text file(s).
    '''
    inp = PhyloNetInput(config)

    tree_ensemble = TreeEnsembleMutPairs(inp)
    mut_pair_dicts = get_mut_pair_dicts(tree_ensemble)
    tree_ensemble.assign_mut_pair_weights(mut_pair_dicts)
    phylo_mut_pairs = tree_ensemble
    
    # Output network in tab-delimited format
    # Source\tTarget\tWeight
    stderr_write(['---------------------'])
    stderr_write(['\nWriting network to file:\n' + inp.net_file]) 
    phylo_mut_pairs.write_network_to_file(inp.net_file)

    # outputs additional information about edge weights
    # to a separate table
    if to_bool(inp.output_edge_weight_table):
        stderr_write(['\nWriting network extended network',
                      'table to file:\n' + inp.net_table_file]) 
        phylo_mut_pairs.write_table_to_file(inp.net_table_file)
 
    # Prints basic stats for reconstructed network
    phylo_mut_pairs.print_stats(aln_pos = inp.pos_list,
                                 pos_subset = inp.pos_subset)


